"""Support for paginated database queries.

This pagination support uses keyset pagination rather than relying on database
cursors, since the latter interact poorly with horizontally scaled services.
"""

import re
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Self, override
from urllib.parse import parse_qs, urlencode

from pydantic import BaseModel
from sqlalchemy import Select, and_, func, or_, select
from sqlalchemy.ext.asyncio import async_scoped_session
from sqlalchemy.orm import InstrumentedAttribute
from starlette.datastructures import URL

from safir.fastapi import ClientRequestError
from safir.models import ErrorLocation

from ._datetime import datetime_to_db

_LINK_REGEX = re.compile(r'\s*<(?P<target>[^>]+)>;\s*rel="(?P<type>[^"]+)"')
"""Matches a component of a valid ``Link`` header."""

__all__ = [
    "CountedPaginatedList",
    "CountedPaginatedQueryRunner",
    "DatetimeIdCursor",
    "InvalidCursorError",
    "PaginatedList",
    "PaginatedQueryRunner",
    "PaginationCursor",
    "PaginationLinkData",
]


class InvalidCursorError(ClientRequestError):
    """The provided cursor was invalid."""

    error = "invalid_cursor"

    def __init__(self, message: str) -> None:
        super().__init__(message, ErrorLocation.query, ["cursor"])


@dataclass
class PaginationLinkData:
    """Holds the data returned in an :rfc:`8288` ``Link`` header."""

    prev_url: str | None
    """URL of the previous page, or `None` for the first page."""

    next_url: str | None
    """URL of the next page, or `None` for the last page."""

    first_url: str | None
    """URL of the first page."""

    @classmethod
    def from_header(cls, header: str | None) -> Self:
        """Parse an :rfc:`8288` ``Link`` with pagination URLs.

        Parameters
        ----------
        header
            Contents of an RFC 8288 ``Link`` header.

        Returns
        -------
        PaginationLinkData
            Parsed form of that header.
        """
        links = {}
        if header:
            for element in header.split(","):
                if m := re.match(_LINK_REGEX, element):
                    if m.group("type") in ("prev", "next", "first"):
                        links[m.group("type")] = m.group("target")
                    elif m.group("type") == "previous":
                        links["prev"] = m.group("target")

        return cls(
            prev_url=links.get("prev"),
            next_url=links.get("next"),
            first_url=links.get("first"),
        )


@dataclass
class PaginationCursor[E: BaseModel](metaclass=ABCMeta):
    """Generic pagnination cursor for keyset pagination.

    The generic type parameter is the Pydantic model into which each row will
    be converted, not the ORM model.
    """

    previous: bool
    """Whether to search backwards instead of forwards."""

    @classmethod
    @abstractmethod
    def from_entry(cls, entry: E, *, reverse: bool = False) -> Self:
        """Construct a cursor with an entry as a bound.

        Builds a cursor to get the entries after the provided entry, or before
        the provided entry if ``reverse`` is set to `True`. When the cursor is
        later applied with `~safir.database.PaginationCursor.apply_cursor`,
        forward cursors (the default) must include the entry the cursor was
        based on. Reverse cursors must exclude the given entry and return data
        starting with the entry immediately previous.

        Parameters
        ----------
        entry
            Basis of the cursor.
        reverse
            Whether to create a previous cursor.

        Returns
        -------
        safir.database.PaginationCursor
            Requested cursor.
        """

    @classmethod
    @abstractmethod
    def from_str(cls, cursor: str) -> Self:
        """Build cursor from the string serialization form.

        Parameters
        ----------
        cursor
            Serialized form of the cursor.

        Returns
        -------
        safir.database.PaginationCursor
            The cursor represented as an object.

        Raises
        ------
        safir.database.InvalidCursorError
            Raised if the cursor is invalid.
        """

    @classmethod
    @abstractmethod
    def apply_order(cls, stmt: Select, *, reverse: bool = False) -> Select:
        """Apply the sort order of the cursor to a select statement.

        This is independent of the cursor and only needs to know the
        underlying ORM fields, so it is available as a class method on the
        cursor class, allowing it to be used without a cursor (such as for the
        initial query). This does, however, mean that the caller has to
        explicitly say whether to reverse the order, which is required when
        using a previous cursor.

        Parameters
        ----------
        stmt
            SQL select statement.
        reverse
            Whether to reverse the sort order.

        Returns
        -------
        sqlalchemy.sql.expression.Select
            The same select statement but sorted in the order expected by the
            cursor.
        """

    @abstractmethod
    def apply_cursor(self, stmt: Select) -> Select:
        """Apply the restrictions from the cursor to a select statement.

        Forward cursors (the default) must include the entry the cursor was
        based on. Reverse cursors must exclude that entry and return data
        beginning with the entry immediately previous.

        Parameters
        ----------
        stmt
            Select statement to modify.

        Returns
        -------
        sqlalchemy.sql.expression.Select
            Modified select statement.
        """

    @abstractmethod
    def invert(self) -> Self:
        """Return the inverted cursor (going the opposite direction).

        Parameters
        ----------
        cursor
            Cursor to invert.

        Returns
        -------
        safir.database.PaginationCursor
            The inverted cursor.
        """


@dataclass
class DatetimeIdCursor[E: BaseModel](PaginationCursor[E], metaclass=ABCMeta):
    """Pagination cursor using a `~datetime.datetime` and unique column ID.

    Cursors that first order by time and then by a unique integer column ID
    can subclass this class and only define the ``id_column`` and
    ``time_column`` static methods to return the ORM model fields for the
    timestamp and column ID.

    Examples
    --------
    Here is a specialization of this cursor class for a simple ORM model where
    the timestamp field to order by is named ``creation_time`` and the unique
    row ID is named ``id``.

    .. code-block:: python

       class TableCursor(DatetimeIdCursor):
           @staticmethod
           def id_column() -> InstrumentedAttribute:
               return Table.id

           @staticmethod
           def time_column() -> InstrumentedAttribute:
               return Table.creation_time
    """

    time: datetime
    """Time position."""

    id: int
    """Unique ID position."""

    @staticmethod
    @abstractmethod
    def id_column() -> InstrumentedAttribute:
        """Return SQL model attribute holding the ID."""

    @staticmethod
    @abstractmethod
    def time_column() -> InstrumentedAttribute:
        """Return SQL model attribute holding the time position."""

    @override
    @classmethod
    def from_str(cls, cursor: str) -> Self:
        previous = cursor.startswith("p")
        if previous:
            cursor = cursor[1:]
        try:
            time, id = cursor.split("_")
            return cls(
                time=datetime.fromtimestamp(float(time), tz=UTC),
                id=int(id),
                previous=previous,
            )
        except Exception as e:
            raise InvalidCursorError(f"Cannot parse cursor: {e!s}") from e

    @override
    @classmethod
    def apply_order(cls, stmt: Select, *, reverse: bool = False) -> Select:
        if reverse:
            return stmt.order_by(cls.time_column(), cls.id_column())
        else:
            return stmt.order_by(
                cls.time_column().desc(), cls.id_column().desc()
            )

    @override
    def apply_cursor(self, stmt: Select) -> Select:
        time = datetime_to_db(self.time)
        time_column = self.time_column()
        id_column = self.id_column()
        if self.previous:
            return stmt.where(
                or_(
                    time_column > time,
                    and_(time_column == time, id_column > self.id),
                )
            )
        else:
            return stmt.where(
                or_(
                    time_column < time,
                    and_(time_column == time, id_column <= self.id),
                )
            )

    @override
    def invert(self) -> Self:
        return type(self)(
            time=self.time, id=self.id, previous=not self.previous
        )

    def __str__(self) -> str:
        previous = "p" if self.previous else ""
        timestamp = self.time.timestamp()

        # Remove a trailing .0, but keep the fractional portion if it matters.
        # This is entirely unnecessary, but it looks nicer in the common case
        # where the database doesn't store milliseconds.
        if int(timestamp) == timestamp:
            timestamp = int(timestamp)

        return f"{previous}{timestamp!s}_{self.id!s}"


@dataclass
class PaginatedList[E: BaseModel, C: PaginationCursor]:
    """Paginated SQL results with accompanying pagination metadata.

    Holds a paginated list of any Pydantic type with pagination cursors. Can
    hold any type of entry and any type of cursor, but implicitly requires the
    entry type be one that is meaningfully paginated by that type of cursor.
    """

    entries: list[E]
    """A batch of entries."""

    next_cursor: C | None
    """Cursor for the next batch of entries."""

    prev_cursor: C | None
    """Cursor for the previous batch of entries."""

    def first_url(self, current_url: URL) -> str:
        """Construct a URL to the first group of results for this query.

        Parameters
        ----------
        current_url
            The starting URL of the current group of entries.

        Returns
        -------
        str
            URL to the first group of entries for this query.
        """
        return str(current_url.remove_query_params("cursor"))

    def next_url(self, current_url: URL) -> str | None:
        """Construct a URL to the next group of results for this query.

        Parameters
        ----------
        current_url
            The starting URL of the current group of entries.

        Returns
        -------
        str or None
            URL to the next group of entries for this query or `None` if there
            are no further entries.
        """
        if not self.next_cursor:
            return None
        first_url = current_url.remove_query_params("cursor")
        params = parse_qs(first_url.query)
        params["cursor"] = [str(self.next_cursor)]
        return str(first_url.replace(query=urlencode(params, doseq=True)))

    def prev_url(self, current_url: URL) -> str | None:
        """Construct a URL to the previous group of results for this query.

        Parameters
        ----------
        current_url
            The starting URL of the current group of entries.

        Returns
        -------
        str or None
            URL to the previous group of entries for this query or `None` if
            there are no further entries.
        """
        if not self.prev_cursor:
            return None
        first_url = current_url.remove_query_params("cursor")
        params = parse_qs(first_url.query)
        params["cursor"] = [str(self.prev_cursor)]
        return str(first_url.replace(query=urlencode(params, doseq=True)))

    def link_header(self, current_url: URL) -> str:
        """Construct an RFC 8288 ``Link`` header for a paginated result.

        Parameters
        ----------
        current_url
            The starting URL of the current group of entries.

        Returns
        -------
        str
            Contents of an RFC 8288 ``Link`` header.
        """
        first_url = self.first_url(current_url)
        next_url = self.next_url(current_url)
        prev_url = self.prev_url(current_url)
        header = f'<{first_url!s}>; rel="first"'
        if next_url:
            header += f', <{next_url!s}>; rel="next"'
        if prev_url:
            header += f', <{prev_url!s}>; rel="prev"'
        return header


class PaginatedQueryRunner[E: BaseModel, C: PaginationCursor]:
    """Run database queries that return paginated results.

    This class implements the logic for keyset pagination based on arbitrary
    SQLAlchemy ORM where clauses.

    Parameters
    ----------
    entry_type
        Type of each entry returned by the queries. This must be a Pydantic
        model.
    cursor_type
        Type of the pagination cursor, which encapsulates the logic of how
        entries are sorted and what set of keys is used to retrieve the next
        or previous batch of entries.
    """

    def __init__(self, entry_type: type[E], cursor_type: type[C]) -> None:
        self._entry_type = entry_type
        self._cursor_type = cursor_type

    async def query_count(
        self, session: async_scoped_session, stmt: Select[tuple]
    ) -> int:
        """Count the number of objects that match a query.

        There is nothing particular to pagination about this query, but it is
        often used in conjunction with pagination to provide the total count
        of matching entries, often in an ``X-Total-Count`` HTTP header.

        Parameters
        ----------
        session
            Database session within which to run the query.
        stmt
            Select statement to execute.

        Returns
        -------
        int
            Count of matching rows.
        """
        count_stmt = select(func.count()).select_from(stmt.subquery())
        return await session.scalar(count_stmt) or 0

    async def query_object(
        self,
        session: async_scoped_session,
        stmt: Select[tuple],
        *,
        cursor: C | None = None,
        limit: int | None = None,
    ) -> PaginatedList[E, C]:
        """Perform a query for objects with an optional cursor and limit.

        Perform the query provided in ``stmt`` with appropriate sorting and
        pagination as determined by the cursor type.

        This method should be used with queries that return a single
        SQLAlchemy model. The provided query will be run with the session
        `~sqlalchemy.ext.asyncio.async_scoped_session.scalars` method and the
        resulting object passed to Pydantic's ``model_validate`` to convert to
        ``entry_type``. For queries returning a tuple of attributes, use
        `query_row` instead.

        Unfortunately, this distinction cannot be type-checked, so be careful
        to use the correct method.

        Parameters
        ----------
        session
            Database session within which to run the query.
        stmt
            Select statement to execute. Pagination and ordering will be
            added, so this statement should not already have limits or order
            clauses applied. This statement must return a list of SQLAlchemy
            ORM models that can be converted to ``entry_type`` by Pydantic.
        cursor
            If present, continue from the provided keyset cursor.
        limit
            If present, limit the result count to at most this number of rows.

        Returns
        -------
        PaginatedList
            Results of the query wrapped with pagination information.
        """
        if cursor or limit:
            return await self._paginated_query(
                session, stmt, cursor=cursor, limit=limit, scalar=True
            )
        else:
            return await self._full_query(session, stmt, scalar=True)

    async def query_row(
        self,
        session: async_scoped_session,
        stmt: Select[tuple],
        *,
        cursor: C | None = None,
        limit: int | None = None,
    ) -> PaginatedList[E, C]:
        """Perform a query for attributes with an optional cursor and limit.

        Perform the query provided in ``stmt`` with appropriate sorting and
        pagination as determined by the cursor type.

        This method should be used with queries that return a list of
        attributes that can be converted to the ``entry_type`` Pydantic model.
        For queries returning a single ORM object, use `query_object` instead.

        Unfortunately, this distinction cannot be type-checked, so be careful
        to use the correct method.

        Parameters
        ----------
        session
            Database session within which to run the query.
        stmt
            Select statement to execute. Pagination and ordering will be
            added, so this statement should not already have limits or order
            clauses applied. This statement must return a tuple of attributes
            that can be converted to ``entry_type`` by Pydantic's
            ``model_validate``.
        cursor
            If present, continue from the provided keyset cursor.
        limit
            If present, limit the result count to at most this number of rows.

        Returns
        -------
        PaginatedList
            Results of the query wrapped with pagination information.
        """
        if cursor or limit:
            return await self._paginated_query(
                session, stmt, cursor=cursor, limit=limit
            )
        else:
            return await self._full_query(session, stmt)

    async def _full_query(
        self,
        session: async_scoped_session,
        stmt: Select[tuple],
        *,
        scalar: bool = False,
    ) -> PaginatedList[E, C]:
        """Perform a full, unpaginated query.

        Parameters
        ----------
        session
            Database session within which to run the query.
        stmt
            Select statement to execute. Pagination and ordering will be
            added, so this statement should not already have limits or order
            clauses applied.
        scalar
            If `True`, the query returns one ORM object for each row instead
            of a tuple of columns.

        Returns
        -------
        PaginatedList
            Results of the query wrapped with pagination information.
        """
        stmt = self._cursor_type.apply_order(stmt)
        if scalar:
            result = await session.scalars(stmt)
        else:
            result = await session.execute(stmt)
        entries = [
            self._entry_type.model_validate(r, from_attributes=True)
            for r in result.all()
        ]
        return PaginatedList[E, C](
            entries=entries, prev_cursor=None, next_cursor=None
        )

    async def _paginated_query(
        self,
        session: async_scoped_session,
        stmt: Select[tuple],
        *,
        cursor: C | None = None,
        limit: int | None = None,
        scalar: bool = False,
    ) -> PaginatedList[E, C]:
        """Perform a paginated query.

        The internal implementation details of the complicated case for
        `query`, where either a cursor or a limit is in play.

        Parameters
        ----------
        session
            Database session within which to run the query.
        stmt
            Select statement to execute. Pagination and ordering will be
            added, so this statement should not already have limits or order
            clauses applied.
        cursor
            If present, continue from the provided keyset cursor.
        limit
            If present, limit the result count to at most this number of rows.
        scalar
            If `True`, the query returns one ORM object for each row instead
            of a tuple of columns.

        Returns
        -------
        PaginatedList
            Results of the query wrapped with pagination information.
        """
        limited_stmt = stmt

        # Apply the cursor, if there is one.
        if cursor:
            limited_stmt = cursor.apply_cursor(limited_stmt)

        # When retrieving a previous set of results using a previous cursor,
        # we have to reverse the sort algorithm so that the cursor boundary
        # can be applied correctly. We'll then later reverse the result set to
        # return it in proper forward-sorted order.
        if cursor and cursor.previous:
            limited_stmt = cursor.apply_order(limited_stmt, reverse=True)
        else:
            limited_stmt = self._cursor_type.apply_order(limited_stmt)

        # Grab one more element than the query limit so that we know whether
        # to create a cursor (because there are more elements) and what the
        # cursor value should be (for forward cursors).
        if limit:
            limited_stmt = limited_stmt.limit(limit + 1)

        # Execute the query twice, once to get the next bach of results and
        # once to get the count of all entries without pagination.
        if scalar:
            result = await session.scalars(limited_stmt)
        else:
            result = await session.execute(limited_stmt)
        entries = [
            self._entry_type.model_validate(r, from_attributes=True)
            for r in result.all()
        ]

        # Calculate the cursors and remove the extra element we asked for.
        prev_cursor = None
        next_cursor = None
        if cursor and cursor.previous:
            next_cursor = cursor.invert()
            if limit and len(entries) > limit:
                prev = entries[limit - 1]
                prev_cursor = self._cursor_type.from_entry(prev, reverse=True)
                entries = entries[:limit]

            # Reverse the results again if we did a reverse sort because we
            # were using a previous cursor.
            entries.reverse()
        else:
            if cursor:
                prev_cursor = cursor.invert()
            if limit and len(entries) > limit:
                next_cursor = self._cursor_type.from_entry(entries[limit])
                entries = entries[:limit]

        # Return the results.
        return PaginatedList[E, C](
            entries=entries, prev_cursor=prev_cursor, next_cursor=next_cursor
        )


@dataclass
class CountedPaginatedList[E: BaseModel, C: PaginationCursor](
    PaginatedList[E, C]
):
    """Paginated SQL results with pagination metadata and total count.

    Holds a paginated list of any Pydantic type, complete with a count and
    cursors. Can hold any type of entry and any type of cursor, but implicitly
    requires the entry type be one that is meaningfully paginated by that type
    of cursor.
    """

    count: int
    """Total number of entries if queried without pagination."""


class CountedPaginatedQueryRunner[E: BaseModel, C: PaginationCursor](
    PaginatedQueryRunner[E, C]
):
    """Run database queries that return paginated results with counts.

    This variation of `PaginatedQueryRunner` always runs a second query to
    count the total number of available entries if queried without pagination.
    It should only be used on small tables or with queries that can be
    satisfied from the table indices; otherwise, the count query could be
    undesirably slow.

    Parameters
    ----------
    entry_type
        Type of each entry returned by the queries. This must be a Pydantic
        model.
    cursor_type
        Type of the pagination cursor, which encapsulates the logic of how
        entries are sorted and what set of keys is used to retrieve the next
        or previous batch of entries.
    """

    @override
    async def query_object(
        self,
        session: async_scoped_session,
        stmt: Select[tuple],
        *,
        cursor: C | None = None,
        limit: int | None = None,
    ) -> CountedPaginatedList[E, C]:
        """Perform a query for objects with an optional cursor and limit.

        Perform the query provided in ``stmt`` with appropriate sorting and
        pagination as determined by the cursor type. Also performs a second
        query to get the total count of entries if retrieved without
        pagination.

        This method should be used with queries that return a single
        SQLAlchemy model. The provided query will be run with the session
        `~sqlalchemy.ext.asyncio.async_scoped_session.scalars` method and the
        resulting object passed to Pydantic's ``model_validate`` to convert to
        ``entry_type``. For queries returning a tuple of attributes, use
        `query_row` instead.

        Unfortunately, this distinction cannot be type-checked, so be careful
        to use the correct method.

        Parameters
        ----------
        session
            Database session within which to run the query.
        stmt
            Select statement to execute. Pagination and ordering will be
            added, so this statement should not already have limits or order
            clauses applied. This statement must return a list of SQLAlchemy
            ORM models that can be converted to ``entry_type`` by Pydantic.
        cursor
            If present, continue from the provided keyset cursor.
        limit
            If present, limit the result count to at most this number of rows.

        Returns
        -------
        CountedPaginatedList
            Results of the query wrapped with pagination information and a
            count of the total number of entries.
        """
        result = await super().query_object(
            session, stmt, cursor=cursor, limit=limit
        )
        count = await self.query_count(session, stmt)
        return CountedPaginatedList[E, C](
            entries=result.entries,
            next_cursor=result.next_cursor,
            prev_cursor=result.prev_cursor,
            count=count,
        )

    @override
    async def query_row(
        self,
        session: async_scoped_session,
        stmt: Select[tuple],
        *,
        cursor: C | None = None,
        limit: int | None = None,
    ) -> CountedPaginatedList[E, C]:
        """Perform a query for attributes with an optional cursor and limit.

        Perform the query provided in ``stmt`` with appropriate sorting and
        pagination as determined by the cursor type. Also performs a second
        query to get the total count of entries if retrieved without
        pagination.

        This method should be used with queries that return a list of
        attributes that can be converted to the ``entry_type`` Pydantic model.
        For queries returning a single ORM object, use `query_object` instead.

        Unfortunately, this distinction cannot be type-checked, so be careful
        to use the correct method.

        Parameters
        ----------
        session
            Database session within which to run the query.
        stmt
            Select statement to execute. Pagination and ordering will be
            added, so this statement should not already have limits or order
            clauses applied. This statement must return a tuple of attributes
            that can be converted to ``entry_type`` by Pydantic's
            ``model_validate``.
        cursor
            If present, continue from the provided keyset cursor.
        limit
            If present, limit the result count to at most this number of rows.

        Returns
        -------
        CountedPaginatedList
            Results of the query wrapped with pagination information and a
            count of the total number of entries.
        """
        result = await super().query_row(
            session, stmt, cursor=cursor, limit=limit
        )
        count = await self.query_count(session, stmt)
        return CountedPaginatedList[E, C](
            entries=result.entries,
            next_cursor=result.next_cursor,
            prev_cursor=result.prev_cursor,
            count=count,
        )



"""Test for long polling when retrieving jobs."""

from __future__ import annotations

import asyncio
from datetime import timedelta

import pytest
from httpx import AsyncClient
from vo_models.uws import JobSummary

from safir.arq.uws import WorkerResult
from safir.datetime import current_datetime, isodatetime
from safir.testing.uws import MockUWSJobRunner, assert_job_summary_equal
from safir.uws._dependencies import UWSFactory

from ..support.uws import SimpleParameters, SimpleXmlParameters

PENDING_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>{}</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Naomi</uws:parameter>
  </uws:parameters>
</uws:job>
"""

EXECUTING_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>EXECUTING</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Naomi</uws:parameter>
  </uws:parameters>
</uws:job>
"""

FINISHED_JOB = """
<uws:job
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobId>1</uws:jobId>
  <uws:ownerId>test-user</uws:ownerId>
  <uws:phase>COMPLETED</uws:phase>
  <uws:creationTime>{}</uws:creationTime>
  <uws:startTime>{}</uws:startTime>
  <uws:endTime>{}</uws:endTime>
  <uws:executionDuration>600</uws:executionDuration>
  <uws:destruction>{}</uws:destruction>
  <uws:parameters>
    <uws:parameter id="name">Naomi</uws:parameter>
  </uws:parameters>
  <uws:results>
    <uws:result id="cutout" xlink:href="https://example.com/some/path"\
 mime-type="application/fits"/>
  </uws:results>
</uws:job>
"""


@pytest.mark.asyncio
async def test_poll(
    client: AsyncClient,
    test_token: str,
    runner: MockUWSJobRunner,
    uws_factory: UWSFactory,
) -> None:
    job_service = uws_factory.create_job_service()
    job = await job_service.create(test_token, SimpleParameters(name="Naomi"))

    # Poll for changes for one second. Nothing will happen since nothing is
    # changing the mock arq queue.
    now = current_datetime()
    r = await client.get("/test/jobs/1", params={"WAIT": "1"})
    assert (current_datetime() - now).total_seconds() >= 1
    assert r.status_code == 200
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        PENDING_JOB.format(
            "PENDING",
            isodatetime(job.creation_time),
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )

    # Start the job and worker.
    r = await client.post(
        "/test/jobs/1/phase", data={"PHASE": "RUN"}, follow_redirects=True
    )
    assert r.status_code == 200
    assert r.url == "https://example.com/test/jobs/1"
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        PENDING_JOB.format(
            "QUEUED",
            isodatetime(job.creation_time),
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )

    # Poll for a change from queued, which we should see after half a second.
    now = current_datetime()
    job, r = await asyncio.gather(
        runner.mark_in_progress(test_token, "1", delay=0.5),
        client.get("/test/jobs/1", params={"WAIT": "2", "phase": "QUEUED"}),
    )
    assert r.status_code == 200
    assert job.start_time
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        EXECUTING_JOB.format(
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )

    # Now, wait again, in parallel with the job finishing. We should get a
    # reply after a second and a half when the job finishes.
    results = [
        WorkerResult(
            result_id="cutout",
            url="s3://some-bucket/some/path",
            mime_type="application/fits",
        )
    ]
    job, r = await asyncio.gather(
        runner.mark_complete(test_token, "1", results, delay=1.5),
        client.get("/test/jobs/1", params={"WAIT": "2", "phase": "EXECUTING"}),
    )
    assert r.status_code == 200
    assert job.start_time
    assert job.end_time
    assert_job_summary_equal(
        JobSummary[SimpleXmlParameters],
        r.text,
        FINISHED_JOB.format(
            isodatetime(job.creation_time),
            isodatetime(job.start_time),
            isodatetime(job.end_time),
            isodatetime(job.creation_time + timedelta(seconds=24 * 60 * 60)),
        ),
    )
    assert (current_datetime() - now).total_seconds() >= 2



"""Camel-case attribute support for Pydantic models."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, override

from pydantic import BaseModel, ConfigDict

__all__ = [
    "CamelCaseModel",
    "to_camel_case",
]


def to_camel_case(string: str) -> str:
    """Convert a string to camel case.

    Intended for use with Pydantic as an alias generator so that the model can
    be initialized from camel-case input, such as Kubernetes objects or
    settings from Helm charts.

    Parameters
    ----------
    string
        Input string.

    Returns
    -------
    str
        String converted to camel-case with the first character in lowercase.

    Examples
    --------
    To support ``camelCase`` input to a model, use the following settings:

    .. code-block:: python

       class Model(BaseModel):
           model_config = ConfigDict(
               alias_generator=to_camel_case, populate_by_name=True
           )

           some_field: str

    This must be added to every class that uses ``snake_case`` for an
    attribute and that needs to be initialized from ``camelCase``.
    Alternately, inherit from `~safir.pydantic.CamelCaseModel`, which is
    derived from `pydantic.BaseModel` with those settings added.
    """
    components = string.split("_")
    return components[0] + "".join(c.title() for c in components[1:])


def _copy_type[**P, T](
    parent: Callable[P, T],
) -> Callable[[Callable[..., T]], Callable[P, T]]:
    """Copy the type of a parent method.

    Used to avoid duplicating the prototype of `pydantic.BaseModel.dict` and
    `pydantic.BaseModel.json` when overriding them in `CamelCaseModel`.

    Parameters
    ----------
    parent
        Method from which to copy a type signature.

    Returns
    -------
    Callable
        Decorator that will replace the type signature of a function with the
        type signature of its parent.
    """
    return lambda x: x


class CamelCaseModel(BaseModel):
    """`pydantic.BaseModel` configured to accept camel-case input.

    This is a convenience class identical to `~pydantic.BaseModel` except with
    an alias generator configured so that it can be initialized with either
    camel-case or snake-case keys. Model exports with ``model_dump`` or
    ``model_dump_json`` also default to exporting in camel-case.
    """

    model_config = ConfigDict(
        alias_generator=to_camel_case, populate_by_name=True
    )

    @override
    @_copy_type(BaseModel.model_dump)
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Export the model as a dictionary.

        Overridden to change the default of ``by_alias`` from `False` to
        `True`, so that by default the exported dictionary uses camel-case.
        """
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        return super().model_dump(**kwargs)

    @override
    @_copy_type(BaseModel.model_dump_json)
    def model_dump_json(self, **kwargs: Any) -> str:
        """Export the model as JSON.

        Overridden to change the default of ``by_alias`` from `False` to
        `True`, so that by default the exported dictionary uses camel-case.
        """
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        return super().model_dump_json(**kwargs)



"""Utility functions for managing an external Uvicorn test process."""

from ._spawn import ServerNotListeningError, UvicornProcess, spawn_uvicorn

__all__ = [
    "ServerNotListeningError",
    "UvicornProcess",
    "spawn_uvicorn",
]



"""Standardized metadata models."""

from __future__ import annotations

from pydantic import BaseModel, Field

__all__ = ["Metadata"]


class Metadata(BaseModel):
    """Metadata about a package."""

    name: str = Field(..., title="Application name", examples=["myapp"])

    version: str = Field(..., title="Version", examples=["1.0.0"])

    description: str | None = Field(
        None, title="Description", examples=["Some package description"]
    )

    repository_url: str | None = Field(
        None, title="Repository URL", examples=["https://example.com/"]
    )

    documentation_url: str | None = Field(
        None, title="Documentation URL", examples=["https://example.com/"]
    )



"""Retrieval of job results.

Job results are stored in a Google Cloud Storage bucket, but UWS requires they
be returned to the user as a URL. This translation layer converts the ``s3``
URL to a signed URL suitable for returning to a client of the service.
"""

from __future__ import annotations

from safir.gcs import SignedURLService

from ._config import UWSConfig
from ._models import JobResult, SignedJobResult

__all__ = ["ResultStore"]


class ResultStore:
    """Result storage handling.

    Parameters
    ----------
    config
        The UWS configuration.
    """

    def __init__(self, config: UWSConfig) -> None:
        self._config = config
        self._url_service = SignedURLService(
            service_account=config.signing_service_account,
            lifetime=config.url_lifetime,
        )

    def sign_url(self, result: JobResult) -> SignedJobResult:
        """Convert a job result into a signed URL.

        Parameters
        ----------
        result
            Result with the URL from the backend.

        Returns
        -------
        SignedJobResult
            Result with any GCS URL replaced with a signed URL.

        Notes
        -----
        This uses custom credentials so that it will work with a GKE service
        account without having to export the secret key as a JSON blob and
        manage it as a secret. For more information, see `gcs_signedurl
        <https://github.com/salrashid123/gcs_signedurl>`__.

        This is probably too inefficient, since it gets new signing
        credentials each time it generates a signed URL. Doing better will
        require figuring out the lifetime and refreshing the credentials when
        the lifetime has expired, which in turn will probably require a
        longer-lived object to hold the credentials.
        """
        if result.url.startswith(("http", "https")):
            # Assume URL is already signed or otherwise directly usable.
            signed_url = result.url
        else:
            mime_type = result.mime_type
            signed_url = self._url_service.signed_url(result.url, mime_type)
        return SignedJobResult(
            id=result.id,
            url=signed_url,
            size=result.size,
            mime_type=result.mime_type,
        )



"""Models for representing metrics events."""

from typing import Any

from dataclasses_avroschema.pydantic import AvroBaseModel
from pydantic import UUID4, AwareDatetime, Field, create_model

from safir.metrics._exceptions import UnsupportedAvroSchemaError

__all__ = ["EventMetadata", "EventPayload"]


class EventMetadata(AvroBaseModel):
    """Common fields for all metrics events.

    Contains the minimum required fields. This gets mixed in
    to a class also containing event payload fields, and
    then gets shipped to Kafka, by the ``EventManager``
    """

    id: UUID4 = Field(
        title="id",
        description="A globally unique value that identifies this event",
    )

    application: str = Field(
        description="The application generating this event.",
        examples=["gafaelfawr", "mobu"],
    )

    timestamp: AwareDatetime = Field(
        description=(
            "The time at which this event occurred, or the time at which this"
            " event completed if it is a duration event."
        ),
    )

    timestamp_ns: int = Field(
        description=(
            "The number of nanoseconds since the unix epoch. This is used as"
            " the InfluxDB timestamp."
        ),
    )


class EventPayload(AvroBaseModel):
    """All event payloads should inherit from this."""

    @classmethod
    def validate_structure(cls) -> None:
        """Do runtime validation of fields.

        Make sure all of the fields are compatible with the backing datastore
        (InfluxDB at the moment).
        """
        valids = [
            "boolean",
            "double",
            "enum",
            "float",
            "int",
            "long",
            "null",
            "string",
        ]
        errors = []

        # Calling cls.avro_schema_to_python() sets a _metadata class variable
        # on this class that memoizes the metadata that comes from any inner
        # Meta class:
        # (https://marcosschroh.github.io/dataclasses-avroschema/records/?h=meta#class-meta)
        #
        # We don't have a Meta class on this payload, but later, we mix this
        # payload into another class that DOES have an inner Meta class that
        # contains the schema name and namespace. If _metadata is already set
        # on this class, that later class will not get that metadata when we
        # eventually call avro_schema_to_python on it.
        #
        # This makes a throw-away class where _metadata will get set, which
        # means the metadata will be generated fresh on the big mixed-in class
        # when we need it.
        tmp_cls = create_model("Temp", __base__=cls)

        schema = tmp_cls.avro_schema_to_python()
        for field in schema["fields"]:
            field_type = field["type"]
            name = field["name"]

            # Unions are represented by a list
            if isinstance(field_type, list):
                if not all(
                    cls._extract_type(subtype) in valids
                    for subtype in field_type
                ):
                    errors.append(
                        f"{name}\n   is a union with a type that is"
                        f" unsupported by InfluxDB: {field_type}"
                    )
                continue

            field_type = cls._extract_type(field_type)

            if field_type not in valids:
                errors.append(
                    f"{name}\n   Serializes to an avro type that is"
                    f" unsupported by InfluxDB: {field_type}"
                )

        if errors:
            errors = ["Unsupported Avro Schema", *errors]
            errors.append(
                f"Supported avro types are these (or unions of these):"
                f" {valids}"
            )
            raise UnsupportedAvroSchemaError("\n".join(errors))

    @classmethod
    def _extract_type(cls, spec: str | dict[str, dict[str, Any]]) -> str:
        """Extract avro base type from container type.

        Raises
        ------
        ValueError
           If a base type can not be extracted
        """
        # Some complex types like enums are represented by a dict, not a
        # string.
        avro_type: str | dict = spec
        if isinstance(spec, dict):
            avro_type = spec["type"]
        if not isinstance(avro_type, str):
            # Ignore suggestion to make this a type error to keep consistent
            # with the ValueError in the calling method.
            raise ValueError(f"{spec} is not a supported Avro type")  # noqa: TRY004
        return avro_type



"""Tests for arq utility functions.

Most of the arq support code is tested by testing the FastAPI dependency.
"""

from __future__ import annotations

from pydantic import SecretStr
from pydantic_core import Url

from safir.arq import build_arq_redis_settings


def test_build_arq_redis_settings() -> None:
    url = Url.build(scheme="redis", host="localhost")
    settings = build_arq_redis_settings(url, None)
    assert settings.host == "localhost"
    assert settings.port == 6379
    assert settings.database == 0
    assert settings.password is None

    url = Url.build(scheme="redis", host="example.com", port=7777, path="4")
    settings = build_arq_redis_settings(url, SecretStr("password"))
    assert settings.host == "example.com"
    assert settings.port == 7777
    assert settings.database == 4
    assert settings.password == "password"



"""Mock Google Cloud Storage API for testing."""

from ._mock import (
    MockBlob,
    MockBucket,
    MockStorageClient,
    patch_google_storage,
)

__all__ = [
    "MockBlob",
    "MockBucket",
    "MockStorageClient",
    "patch_google_storage",
]



"""Tests for the `safir.github.webhooks` module."""

from __future__ import annotations

from pathlib import Path

from safir.github.models import (
    GitHubCheckRunStatus,
    GitHubCheckSuiteConclusion,
    GitHubCheckSuiteStatus,
)
from safir.github.webhooks import (
    GitHubAppInstallationEventAction,
    GitHubAppInstallationEventModel,
    GitHubAppInstallationRepositoriesEventModel,
    GitHubCheckRunEventAction,
    GitHubCheckRunEventModel,
    GitHubCheckSuiteEventModel,
    GitHubPullRequestEventAction,
    GitHubPullRequestEventModel,
    GitHubPushEventModel,
)


def read_webhook_data(filename: str) -> str:
    return (
        Path(__file__).parent.joinpath("data/webhooks", filename).read_text()
    )


def test_push_event() -> None:
    """Test parsing a push event webhook payload."""
    data = GitHubPushEventModel.model_validate_json(
        read_webhook_data("push_event.json")
    )

    assert data.ref == "refs/tags/simple-tag"
    assert data.repository.name == "Hello-World"


def test_installation_event() -> None:
    """Test parsing an installation event webhook payload."""
    data = GitHubAppInstallationEventModel.model_validate_json(
        read_webhook_data("installation.json")
    )

    assert data.action == GitHubAppInstallationEventAction.deleted
    assert data.repositories[0].name == "Hello-World"
    assert data.repositories[0].owner_name == "octocat"


def test_installation_repositories_event() -> None:
    """Test parsing an installation_repositories event webhook payload."""
    data = GitHubAppInstallationRepositoriesEventModel.model_validate_json(
        read_webhook_data("installation_repositories.json")
    )

    assert data.action == "added"
    assert data.repositories_added[0].name == "Space"
    assert data.repositories_added[0].owner_name == "Codertocat"


def test_pull_request_event() -> None:
    """Test parsing a pull_request event webhook payload."""
    data = GitHubPullRequestEventModel.model_validate_json(
        read_webhook_data("pull_request_event.json")
    )

    assert data.number == 2
    assert data.action == GitHubPullRequestEventAction.opened
    assert data.pull_request.number == 2
    assert data.pull_request.title == "Update the README with new information."
    expected_sha = "ec26c3e57ca3a959ca5aad62de7213c562f8c821"
    assert data.pull_request.head.sha == expected_sha


def test_check_suite_completed_event() -> None:
    """Test parsing a check_suite completed event webhook payload."""
    data = GitHubCheckSuiteEventModel.model_validate_json(
        read_webhook_data("check_suite_completed.json")
    )

    assert data.action == "completed"
    assert data.check_suite.id == 118578147
    assert data.check_suite.head_branch == "changes"
    assert data.check_suite.head_sha == (
        "ec26c3e57ca3a959ca5aad62de7213c562f8c821"
    )
    assert data.check_suite.status == GitHubCheckSuiteStatus.completed
    assert data.check_suite.conclusion == GitHubCheckSuiteConclusion.success


def test_check_run_created_event() -> None:
    """Test parsing a check_run created event webhook payload."""
    data = GitHubCheckRunEventModel.model_validate_json(
        read_webhook_data("check_run_created.json")
    )

    assert data.action == GitHubCheckRunEventAction.created
    assert data.check_run.id == 128620228
    assert data.check_run.external_id == ""
    assert str(data.check_run.url) == (
        "https://api.github.com/repos/Codertocat/Hello-World"
        "/check-runs/128620228"
    )
    assert str(data.check_run.html_url) == (
        "https://github.com/Codertocat/Hello-World/runs/128620228"
    )
    assert data.check_run.status == GitHubCheckRunStatus.queued
    assert data.check_run.check_suite.id == 118578147



"""Tests for the safir.kafka module."""

import os
from unittest import mock

import pytest
from aiokafka import AIOKafkaConsumer
from aiokafka.admin.client import AIOKafkaAdminClient
from faststream.kafka import KafkaBroker
from pydantic import SecretStr, ValidationError

from safir.kafka import (
    KafkaConnectionSettings,
    SaslMechanism,
    SecurityProtocol,
)
from tests.support.kafka import KafkaStack


async def assert_clients(settings: KafkaConnectionSettings) -> None:
    admin = None
    broker = None
    consumer = None

    try:
        admin = AIOKafkaAdminClient(**settings.to_aiokafka_params())
        await admin.start()
        await admin.list_topics()

        consumer = AIOKafkaConsumer(**settings.to_aiokafka_params())
        await consumer.start()

        broker = KafkaBroker(**settings.to_faststream_params())
        await broker.start()
        result = await broker.ping(timeout=5)
        assert result

    finally:
        if admin:
            await admin.close()
        if consumer:
            await consumer.stop()
        if broker:
            await broker.stop()


@pytest.mark.asyncio
async def test_plaintext(
    monkeypatch: pytest.MonkeyPatch, kafka_stack: KafkaStack
) -> None:
    bootstrap_server = kafka_stack.kafka_container.get_bootstrap_server()
    settings = KafkaConnectionSettings(
        bootstrap_servers=bootstrap_server,
        security_protocol=SecurityProtocol.PLAINTEXT,
    )
    await assert_clients(settings)

    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            "KAFKA_BOOTSTRAP_SERVERS": bootstrap_server,
            "KAFKA_SECURITY_PROTOCOL": "PLAINTEXT",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        settings = KafkaConnectionSettings()
        await assert_clients(settings)


@pytest.mark.asyncio
async def test_sasl_plaintext(
    monkeypatch: pytest.MonkeyPatch, kafka_stack: KafkaStack
) -> None:
    kafka_container = kafka_stack.kafka_container
    bootstrap_server = kafka_container.get_sasl_plaintext_bootstrap_server()
    with pytest.raises(ValidationError):
        KafkaConnectionSettings(
            bootstrap_servers=bootstrap_server,
            security_protocol=SecurityProtocol.SASL_PLAINTEXT,
            sasl_mechanism=SaslMechanism.SCRAM_SHA_512,
            sasl_username="username",
        )

    settings = KafkaConnectionSettings(
        bootstrap_servers=bootstrap_server,
        security_protocol=SecurityProtocol.SASL_PLAINTEXT,
        sasl_username="admin",
        sasl_password=SecretStr("admin"),
        sasl_mechanism=SaslMechanism.SCRAM_SHA_512,
    )

    await assert_clients(settings)

    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            "KAFKA_BOOTSTRAP_SERVERS": bootstrap_server,
            "KAFKA_SECURITY_PROTOCOL": "SASL_PLAINTEXT",
            "KAFKA_SASL_USERNAME": "admin",
            "KAFKA_SASL_PASSWORD": "admin",
            "KAFKA_SASL_MECHANISM": "SCRAM-SHA-512",
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        settings = KafkaConnectionSettings()

        await assert_clients(settings)


@pytest.mark.asyncio
async def test_sasl_ssl(
    monkeypatch: pytest.MonkeyPatch,
    kafka_stack: KafkaStack,
) -> None:
    kafka_container = kafka_stack.kafka_container
    kafka_cert_path = kafka_stack.kafka_cert_path
    cluster_ca_path = kafka_cert_path / "ca.crt"

    bootstrap_server = kafka_container.get_sasl_ssl_bootstrap_server()
    with pytest.raises(ValidationError):
        KafkaConnectionSettings(
            bootstrap_servers=bootstrap_server,
            security_protocol=SecurityProtocol.SASL_SSL,
            sasl_username="admin",
            sasl_password=SecretStr("admin"),
            cluster_ca_path=cluster_ca_path,
        )

    settings = KafkaConnectionSettings(
        bootstrap_servers=bootstrap_server,
        security_protocol=SecurityProtocol.SASL_SSL,
        sasl_username="admin",
        sasl_password=SecretStr("admin"),
        sasl_mechanism=SaslMechanism.SCRAM_SHA_512,
        cluster_ca_path=cluster_ca_path,
    )

    await assert_clients(settings)

    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            "KAFKA_BOOTSTRAP_SERVERS": bootstrap_server,
            "KAFKA_SECURITY_PROTOCOL": "SASL_SSL",
            "KAFKA_SASL_USERNAME": "admin",
            "KAFKA_SASL_PASSWORD": "admin",
            "KAFKA_SASL_MECHANISM": "SCRAM-SHA-512",
            "KAFKA_CLUSTER_CA_PATH": str(cluster_ca_path),
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        settings = KafkaConnectionSettings()

        await assert_clients(settings)


@pytest.mark.asyncio
async def test_ssl(
    monkeypatch: pytest.MonkeyPatch, kafka_stack: KafkaStack
) -> None:
    kafka_container = kafka_stack.kafka_container
    kafka_cert_path = kafka_stack.kafka_cert_path

    cluster_ca_path = kafka_cert_path / "ca.crt"
    cluster_ca_path = kafka_cert_path / "ca.crt"
    client_cert_path = kafka_cert_path / "client.crt"
    client_key_path = kafka_cert_path / "client.key"

    bootstrap_server = kafka_container.get_ssl_bootstrap_server()
    with pytest.raises(ValidationError):
        KafkaConnectionSettings(
            bootstrap_servers=bootstrap_server,
            security_protocol=SecurityProtocol.SSL,
            cluster_ca_path=cluster_ca_path,
            client_cert_path=client_cert_path,
        )

    settings = KafkaConnectionSettings(
        bootstrap_servers=bootstrap_server,
        security_protocol=SecurityProtocol.SSL,
        cluster_ca_path=cluster_ca_path,
        client_cert_path=client_cert_path,
        client_key_path=client_key_path,
    )

    await assert_clients(settings)

    with mock.patch.dict(os.environ, clear=True):
        envvars = {
            "KAFKA_BOOTSTRAP_SERVERS": bootstrap_server,
            "KAFKA_SECURITY_PROTOCOL": "SSL",
            "KAFKA_CLUSTER_CA_PATH": str(cluster_ca_path),
            "KAFKA_CLIENT_CERT_PATH": str(client_cert_path),
            "KAFKA_CLIENT_KEY_PATH": str(client_key_path),
        }
        for k, v in envvars.items():
            monkeypatch.setenv(k, v)
        settings = KafkaConnectionSettings()

        await assert_clients(settings)



"""Sentry helpers."""

from ._exceptions import SentryException, SentryWebException
from ._helpers import (
    before_send_handler,
    duration,
    fingerprint_env_handler,
    sentry_exception_handler,
)

__all__ = [
    "SentryException",
    "SentryWebException",
    "before_send_handler",
    "duration",
    "fingerprint_env_handler",
    "sentry_exception_handler",
]



"""Test the HTTP client FastAPI dependency."""

from __future__ import annotations

from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Annotated

import pytest
import respx
from asgi_lifespan import LifespanManager
from fastapi import Depends, FastAPI
from httpx import ASGITransport, AsyncClient

from safir.dependencies.http_client import http_client_dependency


@pytest.fixture
def non_mocked_hosts() -> list[str]:
    return ["example.com"]


@asynccontextmanager
async def _lifespan(app: FastAPI) -> AsyncGenerator[None]:
    yield
    await http_client_dependency.aclose()


@pytest.mark.asyncio
async def test_http_client(respx_mock: respx.Router) -> None:
    app = FastAPI(lifespan=_lifespan)
    respx_mock.get("https://www.google.com").respond(200)

    @app.get("/")
    async def handler(
        http_client: Annotated[AsyncClient, Depends(http_client_dependency)],
    ) -> dict[str, str]:
        assert isinstance(http_client, AsyncClient)
        await http_client.get("https://www.google.com")
        return {}

    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with LifespanManager(app):
        async with AsyncClient(transport=transport, base_url=base_url) as c:
            r = await c.get("/")
    assert r.status_code == 200



"""Tools for collecting generic metrics for Arq jobs and queues.

This module is importable separately from the rest of the metrics package
because it depends on the ``arq`` package, which is only installed with the
optional ``safir[arq]`` extra.
"""

from datetime import datetime, timedelta
from typing import Annotated, Any

try:
    from arq.connections import RedisSettings, create_pool
    from arq.typing import StartupShutdown
except ImportError as e:
    raise ImportError(
        "The safir.metrics.arq module requires the arq extra. "
        "Install it with `pip install safir[arq]`."
    ) from e
from pydantic import BaseModel, ConfigDict, Field

from safir.datetime._current import current_datetime

from ._event_manager import EventManager
from ._models import EventPayload

__all__ = [
    "ARQ_EVENTS_CONTEXT_KEY",
    "ArqEvents",
    "ArqMetricsError",
    "ArqQueueJobEvent",
    "ArqQueueStatsEvent",
    "initialize_arq_metrics",
    "make_on_job_start",
    "publish_queue_stats",
]


ARQ_EVENTS_CONTEXT_KEY = "_arq_events"


class ArqQueueJobEvent(EventPayload):
    """Metrics for every job run by Arq."""

    time_in_queue: timedelta
    """Time between ideal job start and when it actually starts."""

    queue: str
    """The queue the job was picked from."""


class ArqQueueStatsEvent(EventPayload):
    """Statistics for an Arq queue."""

    num_queued: int
    """The number of jobs that are currently in the queue."""

    queue: str
    """The queue that these stats apply to."""


class ArqEvents:
    """Container for Arq metrics event publishers."""

    async def initialize(self, manager: EventManager) -> None:
        """Create publishers for generic Arq events.

        Parameters
        ----------
        manager
            An initialized `safir.metrics.EventManager`
        """
        self.arq_queue_job_event = await manager.create_publisher(
            "arq_job_run", ArqQueueJobEvent
        )

        self.arq_queue_stats = await manager.create_publisher(
            "arq_queue_stats", ArqQueueStatsEvent
        )


async def initialize_arq_metrics(
    event_manager: EventManager, ctx: dict
) -> None:
    """Create Arq metrics publishers and inject them into the Arq context.

    Parameters
    ----------
    event_manager
        And initialized `safir.metrics.EventManager`
    ctx
        The context that gets passed to the on_startup function. This will be
        mutated to add the ArqEvents instance.
    """
    events = ArqEvents()
    await events.initialize(event_manager)
    ctx[ARQ_EVENTS_CONTEXT_KEY] = events


class ArqMetricsError(Exception):
    """An exception from misuse/misconfiguration of Arq app metrics."""


class ArqMetricsContext(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    events: Annotated[
        ArqEvents, Field(validation_alias=ARQ_EVENTS_CONTEXT_KEY)
    ]
    """An object that contains Arq metrics events."""

    # The score in the job context is the number of milliseconds since the
    # epoch that this job would ideally start executing. Any difference
    # between now (when is when the job actually starts executing) and this
    # score is time waiting in the queue for other jobs to execute.
    #
    # Note there is also an enqueue_time item in the job context, and this
    # is set to the time that the job was placed into the queue. We can't
    # use this for defered jobs and crons becuase the time between then and
    # now includes the time between when the job was enqueued and when we
    # actually want it to run. Score is the time we actually want it to
    # run.
    ideal_start_time: Annotated[datetime, Field(validation_alias="score")]
    """The time that this job should run."""


def make_on_job_start(queue_name: str) -> StartupShutdown:
    """Make a function that publishes an event for every job execution.

    This should be set as, or composed with, your ``on_job_start`` function in
    your Arq ``WorkerSettings`` class. You need to also call
    `safir.metrics.arq.initialize_arq_metrics` in your worker
    ``on_startup`` function.

    Parameters
    ----------
    queue_name
        The name of the queue that this worker will be listening on.
    """

    async def on_job_start(ctx: dict[Any, Any]) -> Any:
        try:
            context = ArqMetricsContext(**ctx)
        except Exception as e:
            msg = (
                "Arq context does not contain expected values. Make sure you "
                "ran initialize_arq_metrics in your worker on_startup "
                "function."
            )
            raise ArqMetricsError(msg) from e

        now = current_datetime(microseconds=True)
        time_in_queue = now - context.ideal_start_time

        event = ArqQueueJobEvent(
            time_in_queue=time_in_queue,
            queue=queue_name,
        )
        await context.events.arq_queue_job_event.publish(event)

    return on_job_start


async def publish_queue_stats(
    queue: str,
    redis_settings: RedisSettings,
    arq_events: ArqEvents,
) -> None:
    """Publish an event containing statistics about an Arq queue.

    Parameters
    ----------
    queue
        The name of the Arq queue.
    redis_settings
        Connection info for the redis instance containing the queue.
    arq_events
        A collection of Arq metrics event publishers.
    """
    redis = await create_pool(redis_settings)
    num_queued = await redis.zcard(queue)
    event = ArqQueueStatsEvent(queue=queue, num_queued=num_queued)
    await arq_events.arq_queue_stats.publish(event)



"""Tests for the Kubernetes support infrastructure.

These are just basic sanity checks that the mocking is working correctly and
the basic calls work. There is no attempt to test all of the supported APIs.
"""

from __future__ import annotations

import asyncio
from typing import Any

import pytest
from kubernetes_asyncio.client import (
    CoreV1Event,
    V1Container,
    V1DeleteOptions,
    V1ObjectMeta,
    V1ObjectReference,
    V1Pod,
    V1PodSpec,
    V1Secret,
)
from kubernetes_asyncio.watch import Watch

from safir.testing.kubernetes import MockKubernetesApi, strip_none


def test_strip_none() -> None:
    pod = V1Pod(
        metadata=V1ObjectMeta(name="foo"),
        spec=V1PodSpec(
            containers=[
                V1Container(
                    name="something",
                    command=["/bin/sleep", "5"],
                    image="docker.example.com/some/image:latest",
                )
            ]
        ),
    )
    assert strip_none(pod.to_dict()) == {
        "metadata": {"name": "foo"},
        "spec": {
            "containers": [
                {
                    "name": "something",
                    "command": ["/bin/sleep", "5"],
                    "image": "docker.example.com/some/image:latest",
                }
            ]
        },
    }


@pytest.mark.asyncio
async def test_mock(mock_kubernetes: MockKubernetesApi) -> None:
    custom: dict[str, Any] = {
        "apiVersion": "gafaelfawr.lsst.io/v1alpha1",
        "kind": "GafaelfawrServiceToken",
        "metadata": {
            "name": "gafaelfawr-secret",
            "namespace": "mobu",
            "generation": 1,
        },
        "spec": {
            "service": "mobu",
            "scopes": ["admin:token"],
        },
    }
    secret = V1Secret(
        api_version="v1",
        kind="Secret",
        data={"token": "bogus"},
        metadata=V1ObjectMeta(name="gafaelfawr-secret", namespace="mobu"),
        type="Opaque",
    )
    await mock_kubernetes.create_namespaced_custom_object(
        "gafaelfawr.lsst.io",
        "v1alpha1",
        custom["metadata"]["namespace"],
        "gafaelfawrservicetokens",
        custom,
    )
    await mock_kubernetes.create_namespaced_secret(
        secret.metadata.namespace, secret
    )

    assert await mock_kubernetes.list_cluster_custom_object(
        "gafaelfawr.lsst.io", "v1alpha1", "gafaelfawrservicetokens"
    ) == {"items": [custom]}
    assert mock_kubernetes.get_all_objects_for_test("Secret") == [secret]

    def error(method: str, *args: Any) -> None:
        assert method == "replace_namespaced_custom_object"
        raise ValueError("some exception")

    mock_kubernetes.error_callback = error
    with pytest.raises(ValueError, match="some exception") as excinfo:
        await mock_kubernetes.replace_namespaced_custom_object(
            "gafaelfawr.lsst.io",
            "v1alpha1",
            custom["metadata"]["namespace"],
            "gafaelfawrservicetokens",
            "gafaelfawr-secret",
            custom,
        )
    assert str(excinfo.value) == "some exception"


async def watch_events(
    mock_kubernetes: MockKubernetesApi,
    namespace: str,
    *,
    resource_version: str | None = None,
) -> list[CoreV1Event]:
    """Watch events, returning when an event with message ``Done`` is seen."""
    method = mock_kubernetes.list_namespaced_event
    watch_args = {
        "namespace": namespace,
        "timeout_seconds": 10,  # Just in case, so tests don't hang
    }
    if resource_version:
        watch_args["resource_version"] = resource_version
    async with Watch().stream(method, **watch_args) as stream:
        seen = []
        async for event in stream:
            assert event["type"] == "ADDED"
            obj = event["raw_object"]
            seen.append(obj)
            if "Done" in obj.get("message"):
                return seen
        return seen


@pytest.mark.asyncio
async def test_mock_events(mock_kubernetes: MockKubernetesApi) -> None:
    watchers = [
        asyncio.create_task(watch_events(mock_kubernetes, "stuff")),
        asyncio.create_task(watch_events(mock_kubernetes, "stuff")),
        asyncio.create_task(watch_events(mock_kubernetes, "stuff")),
    ]
    await asyncio.sleep(0.2)

    # Creating a pod should by default create an event in that namespace.
    pod = V1Pod(
        metadata=V1ObjectMeta(name="foo"),
        spec=V1PodSpec(
            containers=[
                V1Container(
                    name="something",
                    command=["/bin/sleep", "5"],
                    image="docker.example.com/some/image:latest",
                )
            ]
        ),
    )
    await mock_kubernetes.create_namespaced_pod("stuff", pod)
    some_event = CoreV1Event(
        metadata=V1ObjectMeta(name="some-event"),
        message="Some event",
        involved_object=V1ObjectReference(
            kind="Pod", name="foo", namespace="stuff"
        ),
    )
    await mock_kubernetes.create_namespaced_event("stuff", some_event)
    done_event = CoreV1Event(
        metadata=V1ObjectMeta(name="some-event"),
        message="Done",
        involved_object=V1ObjectReference(
            kind="Pod", name="foo", namespace="stuff"
        ),
    )
    await mock_kubernetes.create_namespaced_event("stuff", done_event)

    # All three watches should see the same thing.
    results = await asyncio.gather(*watchers)
    assert len(results) == 3
    assert results[0] == results[1]
    assert results[0] == results[2]
    result = results[0]
    assert result[1] == some_event.to_dict(serialize=True)
    assert result[2] == done_event.to_dict(serialize=True)

    # The first event should have been the pod running event.
    assert result[0]["message"] == "Pod foo started"
    assert result[0]["involvedObject"]["kind"] == "Pod"
    assert result[0]["involvedObject"]["name"] == "foo"
    assert result[0]["involvedObject"]["namespace"] == "stuff"

    # Starting with resource version "1" should skip the first event, since
    # the semantics of a watch are to show any events *after* the provided
    # resource version.
    events = await watch_events(mock_kubernetes, "stuff", resource_version="1")
    assert events == result[1:]

    # Not specifying a resource version should wait for new events but not see
    # any of the existing events. (Tasks do not start immediately, so we have
    # to wait after creating the task to ensure that it starts running and
    # decides what resource version to start waiting at before we post another
    # event.)
    watcher = asyncio.create_task(watch_events(mock_kubernetes, "stuff"))
    await asyncio.sleep(0.1)
    event = CoreV1Event(
        metadata=V1ObjectMeta(name="some-event"),
        message="Done second time",
        involved_object=V1ObjectReference(
            kind="Pod", name="foo", namespace="stuff"
        ),
    )
    await mock_kubernetes.create_namespaced_event("stuff", event)
    events = await watcher
    assert len(events) == 1
    assert events[0]["message"] == "Done second time"


async def watch_pod_events(
    mock_kubernetes: MockKubernetesApi, name: str, namespace: str
) -> list[dict[str, Any]]:
    """Watch pod events, returning when the pod is deleted."""
    method = mock_kubernetes.list_namespaced_pod
    watch_args = {
        "field_selector": f"metadata.name={name}",
        "namespace": namespace,
        "timeout_seconds": 10,  # Just in case, so tests don't hang
    }
    async with Watch(V1Pod).stream(method, **watch_args) as stream:
        seen = []
        async for event in stream:
            seen.append(event)
            assert isinstance(event["object"], V1Pod)
            if event["type"] == "DELETED":
                return seen
        return seen


@pytest.mark.asyncio
async def test_pod_status(mock_kubernetes: MockKubernetesApi) -> None:
    pod = V1Pod(
        metadata=V1ObjectMeta(name="foo"),
        spec=V1PodSpec(
            containers=[
                V1Container(
                    name="something",
                    command=["/bin/sleep", "5"],
                    image="docker.example.com/some/image:latest",
                )
            ]
        ),
    )
    await mock_kubernetes.create_namespaced_pod("stuff", pod)
    status = await mock_kubernetes.read_namespaced_pod_status("foo", "stuff")
    assert status.status.phase == "Running"
    events = await mock_kubernetes.list_namespaced_event("stuff")
    assert len(events.items) == 1

    mock_kubernetes.initial_pod_phase = "Starting"
    pod = V1Pod(
        metadata=V1ObjectMeta(name="foo"),
        spec=V1PodSpec(
            containers=[
                V1Container(
                    name="something",
                    command=["/bin/sleep", "5"],
                    image="docker.example.com/some/image:latest",
                )
            ]
        ),
    )
    await mock_kubernetes.create_namespaced_pod("other", pod)
    status = await mock_kubernetes.read_namespaced_pod_status("foo", "other")
    assert status.status.phase == "Starting"
    events = await mock_kubernetes.list_namespaced_event("other")
    assert events.items == []
    watchers = [
        asyncio.create_task(watch_pod_events(mock_kubernetes, "foo", "other")),
        asyncio.create_task(watch_pod_events(mock_kubernetes, "foo", "other")),
        asyncio.create_task(watch_pod_events(mock_kubernetes, "foo", "other")),
    ]
    await asyncio.sleep(0.1)
    await mock_kubernetes.patch_namespaced_pod_status(
        "foo",
        "other",
        [{"op": "replace", "path": "/status/phase", "value": "Running"}],
    )
    await mock_kubernetes.delete_namespaced_pod(
        "foo",
        "other",
        grace_period_seconds=1,
        body=V1DeleteOptions(grace_period_seconds=1),
    )

    # All three watches should see the same thing.
    results = await asyncio.gather(*watchers)
    assert len(results) == 3
    assert results[0] == results[1]
    assert results[0] == results[2]
    result = results[0]
    assert result[0]["type"] == "MODIFIED"
    assert result[0]["object"].status.phase == "Running"
    assert result[1]["type"] == "DELETED"
    assert result[1]["object"] == result[0]["object"]



"""Exceptions for the Universal Worker Service."""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Self, override

from vo_models.uws.types import ErrorType

from safir.arq.uws import WorkerError, WorkerErrorType
from safir.datetime import format_datetime_for_logging
from safir.slack.blockkit import (
    SlackCodeBlock,
    SlackException,
    SlackMessage,
    SlackTextBlock,
    SlackTextField,
    SlackWebException,
)
from safir.slack.webhook import SlackIgnoredException

from ._models import JobError

__all__ = [
    "DataMissingError",
    "InvalidPhaseError",
    "ParameterError",
    "SyncJobFailedError",
    "SyncJobNoResultsError",
    "SyncJobTimeoutError",
    "TaskError",
    "UWSError",
    "UnknownJobError",
    "UsageError",
    "WobblyError",
]


class UWSError(SlackIgnoredException):
    """An error with an associated error code.

    SODA requires errors be in ``text/plain`` and start with an error code.
    Adopt that as a general representation of errors produced by the UWS
    layer to simplify generating error responses.

    Parameters
    ----------
    error_code
        SODA error code.
    message
        Exception message, which will be the stringification of the exception.
    detail
        Additional detail.
    """

    def __init__(
        self, error_code: str, message: str, detail: str | None = None
    ) -> None:
        super().__init__(message)
        self.error_code = error_code
        self.detail = detail
        self.status_code = 400


class SyncJobFailedError(UWSError):
    """A sync job failed."""

    def __init__(self, error: JobError) -> None:
        super().__init__(error.code, error.message, error.detail)
        self.status_code = 500


class SyncJobNoResultsError(UWSError):
    """A sync job returned no results."""

    def __init__(self) -> None:
        msg = "Job completed but produced no results"
        super().__init__("Error", msg)
        self.status_code = 500


class SyncJobTimeoutError(UWSError):
    """A sync job timed out before it completed."""

    def __init__(self, timeout: timedelta) -> None:
        msg = f"Job did not complete in {timeout.total_seconds()}s"
        super().__init__("Error", msg)
        self.status_code = 500


class TaskError(SlackException):
    """An error occurred during background task processing.

    This exception is constructed based on `~safir.arq.uws.WorkerError`
    exceptions raised by the backend workers. Those exceptions are not used
    directly to avoid making them, and therefore the worker backend code,
    depend on the Slack error reporting code in Safir.

    Attributes
    ----------
    job_id
        UWS job ID, if known.
    started_at
        When the task was started, if known.
    user
        User whose action triggered this exception, for Slack reporting.
    slack_ignore
        Whether to ignore the error for the purposes of Slack reporting.

    Parameters
    ----------
    error_code
        DALI-compatible error code.
    error_type
        Whether the error is transient or permanent.
    message
        Human-readable error message.
    detail
        Additional details about the error.
    cause_type
        Type of the causing exception, if one is available.
    traceback
        Traceback, if one is available.
    slack_ignore
        Whether to ignore the error for the purposes of Slack reporting.
    """

    def __init__(
        self,
        *,
        error_code: str,
        error_type: ErrorType,
        message: str,
        detail: str | None = None,
        cause_type: str | None = None,
        traceback: str | None = None,
        slack_ignore: bool = False,
    ) -> None:
        super().__init__(message)
        self.job_id: str | None = None
        self.slack_ignore = slack_ignore
        self.started_at: datetime | None = None
        self._error_code = error_code
        self._error_type = error_type
        self._message = message
        self._detail = detail
        self._cause_type = cause_type
        self._traceback = traceback

    @classmethod
    def from_worker_error(cls, exc: WorkerError) -> Self:
        """Create an exception based on a backend worker error.

        Parameters
        ----------
        exc
            Backend worker exception.

        Returns
        -------
        TaskError
            Corresponding task exception.
        """
        slack_ignore = False
        match exc.error_type:
            case WorkerErrorType.FATAL:
                error_code = "Error"
                error_type = ErrorType.FATAL
            case WorkerErrorType.TRANSIENT:
                error_code = "ServiceUnavailable"
                error_type = ErrorType.TRANSIENT
            case WorkerErrorType.USAGE:
                error_code = "UsageError"
                error_type = ErrorType.FATAL
                slack_ignore = True
        return cls(
            error_code=error_code,
            error_type=error_type,
            message=str(exc),
            detail=exc.detail,
            cause_type=exc.cause_type,
            traceback=exc.traceback,
            slack_ignore=slack_ignore,
        )

    def to_job_error(self) -> JobError:
        """Convert to a `~safir.uws._models.UWSJobError`."""
        if self._traceback and self._detail:
            detail: str | None = self._detail + "\n\n" + self._traceback
        else:
            detail = self._detail or self._traceback
        return JobError(
            code=self._error_code,
            type=self._error_type,
            message=self._message,
            detail=detail,
        )

    @override
    def to_slack(self) -> SlackMessage:
        message = super().to_slack()
        if self._traceback:
            block = SlackCodeBlock(heading="Traceback", code=self._traceback)
            message.attachments.append(block)
        if self.started_at:
            started_at = format_datetime_for_logging(self.started_at)
            field = SlackTextField(heading="Started at", text=started_at)
            message.fields.insert(1, field)
        if self.job_id:
            field = SlackTextField(heading="UWS job ID", text=self.job_id)
            message.fields.insert(1, field)
        if self._cause_type:
            text = SlackTextBlock(
                heading="Original exception", text=self._cause_type
            )
            message.blocks.append(text)
        if self._detail:
            text = SlackTextBlock(heading="Detail", text=self._detail)
            message.blocks.append(text)
        return message


class UsageError(UWSError):
    """Invalid parameters were passed to a UWS API."""

    def __init__(self, message: str, detail: str | None = None) -> None:
        super().__init__("UsageError", message, detail)
        self.status_code = 422


class DataMissingError(UWSError):
    """The data requested does not exist for that job."""

    def __init__(self, message: str) -> None:
        super().__init__("UsageError", message)
        self.status_code = 404


class InvalidPhaseError(UsageError):
    """The job is in an invalid phase for the desired operation."""


class ParameterError(UsageError):
    """Unsupported value passed to a parameter."""


class UnknownJobError(DataMissingError):
    """The named job could not be found in the database."""

    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job {job_id} not found")
        self.job_id = job_id


class WobblyError(SlackWebException):
    """An error occurred making a request to Wobbly."""



"""Test the Gafaelfawr auth FastAPI dependencies."""

from __future__ import annotations

import json
from typing import Annotated
from unittest.mock import ANY

import pytest
from fastapi import Depends, FastAPI
from httpx import ASGITransport, AsyncClient
from structlog.stdlib import BoundLogger

from safir.dependencies.gafaelfawr import (
    auth_delegated_token_dependency,
    auth_dependency,
    auth_logger_dependency,
)
from safir.logging import configure_logging


@pytest.mark.asyncio
async def test_auth_dependency() -> None:
    app = FastAPI()

    @app.get("/")
    async def handler(
        user: Annotated[str, Depends(auth_dependency)],
    ) -> dict[str, str]:
        return {"user": user}

    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/")
        assert r.status_code == 422

        r = await client.get("/", headers={"X-Auth-Request-User": "someuser"})
        assert r.status_code == 200
        assert r.json() == {"user": "someuser"}


@pytest.mark.asyncio
async def test_auth_delegated_token_dependency() -> None:
    app = FastAPI()

    @app.get("/")
    async def handler(
        token: Annotated[str, Depends(auth_delegated_token_dependency)],
    ) -> dict[str, str]:
        return {"token": token}

    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/")
        assert r.status_code == 422

        r = await client.get(
            "/", headers={"X-Auth-Request-Token": "sometoken"}
        )
        assert r.status_code == 200
        assert r.json() == {"token": "sometoken"}


@pytest.mark.asyncio
async def test_auth_logger_dependency(
    caplog: pytest.LogCaptureFixture,
) -> None:
    configure_logging(name="myapp", profile="production", log_level="info")

    app = FastAPI()

    @app.get("/")
    async def handler(
        logger: Annotated[BoundLogger, Depends(auth_logger_dependency)],
    ) -> dict[str, str]:
        logger.info("something")
        return {}

    caplog.clear()
    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/", headers={"User-Agent": ""})
        assert r.status_code == 422

        r = await client.get(
            "/", headers={"User-Agent": "", "X-Auth-Request-User": "someuser"}
        )
        assert r.status_code == 200

    assert len(caplog.record_tuples) == 1
    assert json.loads(caplog.record_tuples[0][2]) == {
        "event": "something",
        "httpRequest": {
            "requestMethod": "GET",
            "requestUrl": "https://example.com/",
            "remoteIp": "127.0.0.1",
        },
        "logger": "myapp",
        "request_id": ANY,
        "severity": "info",
        "user": "someuser",
    }



"""Tests for Sentry helpers."""

import pytest
import respx
import sentry_sdk
from httpx import AsyncClient, HTTPError, Response

from safir.sentry import SentryException, SentryWebException
from safir.testing.sentry import Captured


def test_env_fingerprint_before_send(
    sentry_fingerprint_items: Captured,
) -> None:
    sentry_sdk.capture_exception(Exception("some error"))
    (error,) = sentry_fingerprint_items.errors
    assert error["fingerprint"] == ["{{ default }}", "some_env"]


def test_sentry_exception_before_send(
    sentry_exception_items: Captured,
) -> None:
    class SomeError(SentryException): ...

    exc = SomeError("some error")
    exc.tags["woo"] = "hoo"
    exc.contexts["foo"] = {"bar": "baz"}

    sentry_sdk.capture_exception(exc)

    (error,) = sentry_exception_items.errors
    assert error["contexts"]["foo"] == {"bar": "baz"}
    assert error["tags"] == {"woo": "hoo"}


def test_combined_before_send(sentry_combo_items: Captured) -> None:
    class SomeError(SentryException): ...

    exc = SomeError("some error")
    exc.tags["woo"] = "hoo"
    exc.contexts["foo"] = {"bar": "baz"}

    sentry_sdk.capture_exception(exc)

    (error,) = sentry_combo_items.errors
    assert error["contexts"]["foo"] == {"bar": "baz"}
    assert error["fingerprint"] == ["{{ default }}", "some_env"]
    assert error["tags"] == {"woo": "hoo"}


def test_sentry_exception(sentry_exception_items: Captured) -> None:
    class SomeError(SentryException): ...

    exc = SomeError("some error")
    exc.tags["woo"] = "hoo"
    exc.contexts["foo"] = {"bar": "baz"}

    sentry_sdk.capture_exception(exc)

    (error,) = sentry_exception_items.errors
    assert error["contexts"]["foo"] == {"bar": "baz"}
    assert error["tags"] == {"woo": "hoo"}


@pytest.mark.asyncio
async def test_sentry_web_exception(
    respx_mock: respx.Router, sentry_exception_items: Captured
) -> None:
    class SomeError(SentryWebException):
        pass

    respx_mock.get("https://example.org/").mock(return_value=Response(404))
    exception = None
    try:
        async with AsyncClient() as client:
            r = await client.get("https://example.org/")
            r.raise_for_status()
    except HTTPError as e:
        exception = SomeError.from_exception(e)
        assert str(exception) == "Status 404 from GET https://example.org/"
        sentry_sdk.capture_exception(exception)

    (error,) = sentry_exception_items.errors
    assert error["exception"]["values"][0]["type"] == (
        "test_sentry_web_exception.<locals>.SomeError"
    )
    assert error["exception"]["values"][0]["value"] == (
        "Status 404 from GET https://example.org/"
    )
    assert error["tags"] == {
        "httpx_request_method": "GET",
        "httpx_request_status": "404",
        "httpx_request_url": "https://example.org/",
    }



"""Schema manager exceptions."""

__all__ = [
    "IncompatibleSchemaError",
    "InvalidAvroNameError",
    "InvalidMetadataError",
    "UnknownDeserializeError",
    "UnknownSchemaError",
]


from dataclasses_avroschema.pydantic import AvroBaseModel


class IncompatibleSchemaError(Exception):
    """A schema is incompatible with the latest version in the registry."""


class InvalidAvroNameError(Exception):
    """The decalred name or namespace for an Avro schema is not valid."""


class InvalidMetadataError(Exception):
    """The Meta inner class on a model has unexpected values in fields."""


class UnknownDeserializeError(Exception):
    """The schema registry client returns None when deserializing."""


class UnknownSchemaError(Exception):
    """A schema is not managed by the Registry, and therefore cannot be
    serialized into a native Python object.
    """

    def __init__(self, data: AvroBaseModel, subject: str) -> None:
        self.message = (
            f"Schema for model: {type(data).__name__} with subject: {subject}"
            " is not known to the manager. ``register`` must be called before"
            " you try to serialize instances of this model."
        )
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



"""Utility functions for database management."""

from ._alembic import (
    AlembicConfigError,
    is_database_current,
    run_migrations_offline,
    run_migrations_online,
    stamp_database,
    stamp_database_async,
    unstamp_database,
)
from ._connection import create_async_session, create_database_engine
from ._datetime import datetime_from_db, datetime_to_db
from ._initialize import (
    DatabaseInitializationError,
    drop_database,
    initialize_database,
)
from ._pagination import (
    CountedPaginatedList,
    CountedPaginatedQueryRunner,
    DatetimeIdCursor,
    InvalidCursorError,
    PaginatedList,
    PaginatedQueryRunner,
    PaginationCursor,
    PaginationLinkData,
)
from ._retry import retry_async_transaction

__all__ = [
    "AlembicConfigError",
    "CountedPaginatedList",
    "CountedPaginatedQueryRunner",
    "DatabaseInitializationError",
    "DatetimeIdCursor",
    "InvalidCursorError",
    "PaginatedList",
    "PaginatedQueryRunner",
    "PaginationCursor",
    "PaginationLinkData",
    "create_async_session",
    "create_database_engine",
    "datetime_from_db",
    "datetime_to_db",
    "drop_database",
    "initialize_database",
    "is_database_current",
    "retry_async_transaction",
    "run_migrations_offline",
    "run_migrations_online",
    "stamp_database",
    "stamp_database_async",
    "unstamp_database",
]



"""Dependencies for metrics functionality."""

from abc import ABC, abstractmethod

from safir.metrics import EventManager

__all__ = [
    "EventDependency",
    "EventMaker",
]


class EventMaker(ABC):
    """A blueprint for an event publisher container class."""

    @abstractmethod
    async def initialize(self, manager: EventManager) -> None:
        """Create event publishers and assign to instance attributes.

        Use ``manager.create_publisher`` to assign event publishers to
        instance attributes.

        Parameters
        ----------
        manager
            An ``EventManager`` to create event publishers
        """


class EventDependency[E: EventMaker]:
    """Provides EventManager-managed events for apps to publish.

    Parameters
    ----------
    event_maker
        An instance of an implementation of ``EventMaker``

    Examples
    --------
    .. code-block:: python
       :caption: myapp.dependencies.events_dependency

       from safir.metrics import (
           EventMaker,
           EventManager,
           EventPayload,
           EventsDependency,
           MetricsConfigurationWithKafka,
       )


       class Event1(EventPayload):
           foo: str
           bar: str


       class event2(EventPayload):
           baz: int
           buz: str


       class Events(EventMaker):
           async def initialize(self, manager: EventManager) -> None:
               self.event1 = await manager.create_publisher("event1", Event1)
               self.event2 = await manager.create_publisher("event2", Event2)


       dependency = EventsDependency(AppEvents())

    .. code-block:: python
       :caption: myapp.main

       from myapp.dependencies.events_dependency import dependency as ed

       # In application init, maybe a lifecycle method in a FastAPI app
       manager = MetricsConfigurationWithKafka().make_manager()
       await manager.initialize()
       await events.initialize(manager)

    .. code-block:: python
       :caption: myapp.someservice

       from myapp.dependencies.events_dependency import dependency as ed


       events = ed()

       # In app code
       await events.event1.publish(Event1(foo="foo", bar="bar"))
       await events.event2.publish(Event2(baz=123, buz="buz"))
    """

    def __init__(self, event_maker: E) -> None:
        self._events = event_maker
        self._initialized = False

    async def initialize(
        self,
        manager: EventManager,
    ) -> None:
        """Initialize the event maker and set the attribute.

        Parameters
        ----------
        manager
            An ``EventManager`` to register and publish events.
        """
        await self._events.initialize(manager)
        self._initialized = True

    @property
    def events(self) -> E:
        """Return the instantiated event maker.

        Raises
        ------
        RuntimeError
            If this ``EventsDependency`` hasn't been intialized yet.
        """
        if not self._initialized:
            raise RuntimeError("EventsDependency not initialized")
        return self._events

    async def __call__(self) -> E:
        """Return the initialized event maker for use as a FastAPI
        dependency.
        """
        return self.events



"""Mocks for testing services using the Safir UWS support."""

from __future__ import annotations

import asyncio
import json
from collections import defaultdict
from datetime import UTC, datetime
from urllib.parse import parse_qs

import respx
from httpx import AsyncClient, Request, Response
from vo_models.uws.types import ExecutionPhase

from safir.arq import JobMetadata, JobResult, MockArqQueue
from safir.arq.uws import WorkerResult
from safir.datetime import current_datetime, parse_isodatetime
from safir.uws import (
    Job,
    JobCreate,
    JobUpdateAborted,
    JobUpdateCompleted,
    JobUpdateError,
    JobUpdateExecuting,
    JobUpdateMetadata,
    JobUpdateQueued,
    SerializedJob,
    UWSConfig,
)
from safir.uws._storage import JobStore

__all__ = [
    "MockUWSJobRunner",
    "MockWobbly",
    "patch_wobbly",
]


class MockWobbly:
    """Mock the Wobbly web service, which stores UWS job information.

    Use of this mock web service requires presentation of a token generated by
    the `make_token` class method, which encodes username and service
    information that would normally be taken from HTTP headers after
    Gafaelfawr processing.

    Attributes
    ----------
    jobs
        Stored jobs, organized by service, username, and then job ID.
    """

    @staticmethod
    def make_token(service: str, username: str) -> str:
        """Create a fake internal token for Wobbly calls.

        Parameters
        ----------
        service
            Service name encoded in fake internal token.
        username
            Username encoded in fake internal token.

        Returns
        -------
        str
            Fake internal token.
        """
        return f"gt-{service}.{username}"

    def __init__(self) -> None:
        # Next available job ID.
        self._job_id = 1

        # Maps service to username to job ID to a job record.
        self.jobs: defaultdict[str, defaultdict[str, dict[str, SerializedJob]]]
        self.jobs = defaultdict(lambda: defaultdict(dict))

    def create_job(self, request: Request) -> Response:
        """Create a new job record."""
        service, username = self._get_auth(request)
        job_create = JobCreate.model_validate_json(request.content)
        job_id = str(self._job_id)
        self._job_id += 1
        job = SerializedJob(
            id=job_id,
            service=service,
            owner=username,
            phase=ExecutionPhase.PENDING,
            creation_time=current_datetime(),
            **job_create.model_dump(),
        )
        self.jobs[service][username][job_id] = job
        return Response(
            201,
            json=job.model_dump(mode="json"),
            headers={"Location": str(request.url) + f"/{job_id}"},
        )

    def delete_job(self, request: Request, *, job_id: str) -> Response:
        """Delete a job."""
        service, username = self._get_auth(request)
        if job_id not in self.jobs[service][username]:
            return Response(404)
        del self.jobs[service][username][job_id]
        return Response(204)

    def get_job(self, request: Request, *, job_id: str) -> Response:
        """Retrieve a job."""
        service, username = self._get_auth(request)
        job = self.jobs[service][username].get(job_id)
        if not job:
            return Response(404)
        return Response(200, json=job.model_dump(mode="json"))

    def list_jobs(self, request: Request) -> Response:
        """List jobs matching the search parameters.

        Cursors are not supported. ``limit`` is, but does not result in a
        paginated response.
        """
        service, username = self._get_auth(request)

        # Parse query.
        query = parse_qs(request.url.query.decode())
        phases = set()
        if "phase" in query:
            phases = set(query["phase"])
        since = None
        if "since" in query:
            since = parse_isodatetime(query["since"][0])

        # Perform the search.
        results = []
        for job in self.jobs[service][username].values():
            if phases and job.phase not in phases:
                continue
            if since and job.creation_time <= since:
                continue
            results.append(job)

        # Sort the results and limit them if needed.
        json_results = [
            j.model_dump(mode="json")
            for j in sorted(
                results, key=lambda j: (j.creation_time, j.id), reverse=True
            )
        ]
        if "limit" in query:
            limit = int(query["limit"][0])
            if len(json_results) > limit:
                json_results = json_results[:limit]

        # Return the response.
        return Response(200, json=json_results)

    def update_job(self, request: Request, *, job_id: str) -> Response:
        """Make an update to a job record."""
        service, username = self._get_auth(request)
        job = self.jobs[service][username].get(job_id)
        if not job:
            return Response(404)
        body = json.loads(request.content)

        # First handle the only case without a phase, which is a metadata
        # update.
        if "phase" not in body or not body["phase"]:
            update = JobUpdateMetadata.model_validate(body)
            job.destruction_time = update.destruction_time
            job.execution_duration = update.execution_duration
            return Response(200, json=job.model_dump(mode="json"))

        # Otherwise, handle all the phase modification cases.
        match body["phase"]:
            case ExecutionPhase.ABORTED:
                _ = JobUpdateAborted.model_validate(body)
                if job.start_time:
                    job.end_time = current_datetime()
            case ExecutionPhase.COMPLETED:
                completed_update = JobUpdateCompleted.model_validate(body)
                job.end_time = current_datetime()
                job.results = completed_update.results
            case ExecutionPhase.EXECUTING:
                executing_update = JobUpdateExecuting.model_validate(body)
                start_time = executing_update.start_time.replace(microsecond=0)
                job.start_time = start_time
            case ExecutionPhase.ERROR:
                error_update = JobUpdateError.model_validate(body)
                job.end_time = current_datetime()
                job.errors = error_update.errors
            case ExecutionPhase.QUEUED:
                queued_update = JobUpdateQueued.model_validate(body)
                job.message_id = queued_update.message_id
        job.phase = ExecutionPhase[body["phase"]]
        return Response(200, json=job.model_dump(mode="json"))

    def _get_auth(self, request: Request) -> tuple[str, str]:
        """Parse the fake internal token into service and username."""
        auth_type, token = request.headers["Authorization"].split()
        assert auth_type.lower() == "bearer"
        assert token.startswith("gt-")
        service, username = token[3:].split(".", 1)
        return (service, username)


def patch_wobbly(respx_mock: respx.Router, wobbly_url: str) -> MockWobbly:
    """Set up the mock for a Wobbly server.

    Parameters
    ----------
    respx_mock
        Mock router.
    wobby_url
        Base URL on which Wobbly is "listening."

    Returns
    -------
    MockWobbly
        Mock Wobbly service.
    """
    wobbly_url = wobbly_url.rstrip("/")
    mock = MockWobbly()
    respx_mock.get(wobbly_url + "/jobs").mock(side_effect=mock.list_jobs)
    respx_mock.post(wobbly_url + "/jobs").mock(side_effect=mock.create_job)
    job_url = rf"{wobbly_url}/jobs/(?P<job_id>[^/]+)"
    respx_mock.get(url__regex=job_url).mock(side_effect=mock.get_job)
    respx_mock.delete(url__regex=job_url).mock(side_effect=mock.delete_job)
    respx_mock.patch(url__regex=job_url).mock(side_effect=mock.update_job)
    return mock


class MockUWSJobRunner:
    """Simulate execution of jobs with a mock queue.

    When running the test suite, the arq queue is replaced with a mock queue
    that doesn't execute workers. That execution has to be simulated by
    manually updating state in the mock queue and running the UWS database
    worker functions that normally would be run automatically by the queue.

    Parameters
    ----------
    config
        UWS configuration.
    arq_queue
        Mock arq queue for testing.
    """

    def __init__(self, config: UWSConfig, arq_queue: MockArqQueue) -> None:
        self._arq = arq_queue
        self._store = JobStore(config, AsyncClient())

    async def get_job_metadata(self, token: str, job_id: str) -> JobMetadata:
        """Get the arq job metadata for a job.

        Parameters
        ----------
        token
            Token for the user.
        job_id
            UWS job ID.

        Returns
        -------
        JobMetadata
            arq job metadata.
        """
        job = await self._store.get(token, job_id)
        assert job.message_id
        return await self._arq.get_job_metadata(job.message_id)

    async def get_job_result(self, token: str, job_id: str) -> JobResult:
        """Get the arq job result for a job.

        Parameters
        ----------
        token
            Token for the user.
        job_id
            UWS job ID.

        Returns
        -------
        JobMetadata
            arq job metadata.
        """
        job = await self._store.get(token, job_id)
        assert job.message_id
        return await self._arq.get_job_result(job.message_id)

    async def mark_in_progress(
        self, token: str, job_id: str, *, delay: float | None = None
    ) -> Job:
        """Mark a queued job in progress.

        Parameters
        ----------
        token
            Token for the user.
        job_id
            Job ID.
        delay
            How long to delay in seconds before marking the job as complete.

        Returns
        -------
        Job
            Record of the job.
        """
        if delay:
            await asyncio.sleep(delay)
        job = await self._store.get(token, job_id)
        assert job.message_id
        await self._arq.set_in_progress(job.message_id)
        await self._store.mark_executing(token, job_id, datetime.now(tz=UTC))
        return await self._store.get(token, job_id)

    async def mark_complete(
        self,
        token: str,
        job_id: str,
        results: list[WorkerResult] | Exception,
        *,
        delay: float | None = None,
    ) -> Job:
        """Mark an in progress job as complete.

        Parameters
        ----------
        token
            Token for the user.
        job_id
            Job ID.
        results
            Results to return. May be an exception to simulate a job failure.
        delay
            How long to delay in seconds before marking the job as complete.

        Returns
        -------
        Job
            Record of the job.
        """
        if delay:
            await asyncio.sleep(delay)
        job = await self._store.get(token, job_id)
        assert job.message_id
        await self._arq.set_complete(job.message_id, result=results)
        job_result = await self._arq.get_job_result(job.message_id)
        await self._store.mark_completed(token, job_id, job_result)
        return await self._store.get(token, job_id)



"""Test ``X-Forwarded-For`` middleware."""

from __future__ import annotations

from ipaddress import _BaseNetwork, ip_network

import pytest
from fastapi import FastAPI, Request
from httpx import ASGITransport, AsyncClient
from starlette.datastructures import Headers

from safir.middleware.x_forwarded import XForwardedMiddleware


def build_app(proxies: list[_BaseNetwork] | None = None) -> FastAPI:
    """Construct a test FastAPI app with the middleware registered."""
    app = FastAPI()
    app.add_middleware(XForwardedMiddleware, proxies=proxies)
    return app


@pytest.mark.asyncio
async def test_ok() -> None:
    app = build_app([ip_network("11.0.0.0/8")])

    @app.get("/")
    async def handler(request: Request) -> dict[str, str]:
        assert request.client
        assert request.client.host == "10.10.10.10"
        assert request.state.forwarded_host == "foo.example.com"
        assert request.url == "https://foo.example.com/"
        return {}

    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get(
            "/",
            headers={
                "Host": "foo.example.com",
                "X-Forwarded-For": "10.10.10.10, 11.11.11.11",
                "X-Forwarded-Proto": "https, http",
                "X-Forwarded-Host": "foo.example.com",
            },
        )
    assert r.status_code == 200


@pytest.mark.asyncio
async def test_defaults() -> None:
    app = build_app()

    @app.get("/")
    async def handler(request: Request) -> dict[str, str]:
        assert request.client
        assert request.client.host == "192.168.0.1"
        assert request.state.forwarded_host == "foo.example.com"
        assert request.url == "http://example.com/"
        return {}

    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get(
            "/",
            headers={
                "X-Forwarded-For": ("1.1.1.1, 192.168.0.1"),
                "X-Forwarded-Proto": "https, http",
                "X-Forwarded-Host": "foo.example.com",
            },
        )
    assert r.status_code == 200


@pytest.mark.asyncio
async def test_no_forwards() -> None:
    app = build_app([ip_network("127.0.0.1")])

    @app.get("/")
    async def handler(request: Request) -> dict[str, str]:
        assert not request.state.forwarded_host
        assert request.client
        assert request.client.host == "127.0.0.1"
        assert request.url == "http://example.com/"
        return {}

    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/")
    assert r.status_code == 200


@pytest.mark.asyncio
async def test_all_filtered() -> None:
    app = build_app([ip_network("10.0.0.0/8")])

    @app.get("/")
    async def handler(request: Request) -> dict[str, str]:
        assert request.client
        assert request.client.host == "10.10.10.10"
        assert request.state.forwarded_host == "foo.example.com"
        assert request.url == "https://example.com/"
        return {}

    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get(
            "/",
            headers={
                "X-Forwarded-For": "10.10.10.10, 10.0.0.1",
                "X-Forwarded-Proto": "https, http",
                "X-Forwarded-Host": "foo.example.com",
            },
        )
    assert r.status_code == 200


@pytest.mark.asyncio
async def test_one_proto() -> None:
    app = build_app([ip_network("11.11.11.11")])

    @app.get("/")
    async def handler(request: Request) -> dict[str, str]:
        assert request.client
        assert request.client.host == "10.10.10.10"
        assert request.state.forwarded_host == "foo.example.com"
        assert request.url == "https://example.com/"
        return {}

    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get(
            "/",
            headers={
                "X-Forwarded-For": "10.10.10.10, 11.11.11.11",
                "X-Forwarded-Proto": "https",
                "X-Forwarded-Host": "foo.example.com",
            },
        )
    assert r.status_code == 200


@pytest.mark.asyncio
async def test_no_proto_or_host() -> None:
    app = build_app([ip_network("11.11.11.11")])

    @app.get("/")
    async def handler(request: Request) -> dict[str, str]:
        assert not request.state.forwarded_host
        assert request.client
        assert request.client.host == "10.10.10.10"
        assert request.url == "http://example.com/"
        return {}

    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get(
            "/", headers={"X-Forwarded-For": "10.10.10.10, 11.11.11.11"}
        )
    assert r.status_code == 200


@pytest.mark.asyncio
async def test_too_many_headers() -> None:
    """Test handling of duplicate headers.

    HTTPX doesn't allow passing in duplicate headers, so we cannot test end to
    end.  Instead, test by generating a mock request and then calling the
    underling middleware functions directly.
    """
    scope = {
        "type": "http",
        "headers": [
            ("X-Forwarded-For", "10.10.10.10"),
            ("X-Forwarded-For", "10.10.10.1"),
            ("X-Forwarded-Proto", "https"),
            ("X-Forwarded-Proto", "http"),
            ("X-Forwarded-Host", "example.org"),
            ("X-Forwarded-Host", "example.com"),
        ],
    }
    headers = Headers(scope=scope)
    app = FastAPI()
    middleware = XForwardedMiddleware(app, proxies=[ip_network("10.0.0.0/8")])
    assert middleware._get_forwarded_for(headers) == []
    assert middleware._get_forwarded_proto(headers) == []
    assert not middleware._get_forwarded_host(headers)



"""Models for UWS services.

See https://www.ivoa.net/documents/UWS/20161024/REC-UWS-1.1-20161024.html for
most of these models. Descriptive language here is paraphrased from this
standard.
"""

from abc import ABC, abstractmethod
from typing import Annotated, Any, Literal, Self, override

from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer
from vo_models.uws import (
    ErrorSummary,
    JobSummary,
    Parameters,
    ResultReference,
    Results,
    ShortJobDescription,
)
from vo_models.uws.types import ErrorType, ExecutionPhase, UWSVersion

from safir.arq.uws import WorkerResult
from safir.pydantic import SecondsTimedelta, UtcDatetime

__all__ = [
    "ACTIVE_PHASES",
    "Job",
    "JobCreate",
    "JobError",
    "JobResult",
    "JobUpdateAborted",
    "JobUpdateCompleted",
    "JobUpdateError",
    "JobUpdateExecuting",
    "JobUpdateMetadata",
    "JobUpdateQueued",
    "ParametersModel",
    "SignedJobResult",
]


ACTIVE_PHASES = {
    ExecutionPhase.PENDING,
    ExecutionPhase.QUEUED,
    ExecutionPhase.EXECUTING,
}
"""Phases in which the job is active and can be waited on."""


class ParametersModel[W: BaseModel, X: Parameters](BaseModel, ABC):
    """Defines the interface for a model suitable for job parameters."""

    @abstractmethod
    def to_worker_parameters(self) -> W:
        """Convert to the domain model used by the backend worker."""

    @abstractmethod
    def to_xml_model(self) -> X:
        """Convert to the XML model used in XML API responses."""


class JobError(BaseModel):
    """Failure information about a job."""

    type: Annotated[
        ErrorType,
        Field(
            title="Error type",
            description="Type of the error",
            examples=[ErrorType.TRANSIENT, ErrorType.FATAL],
        ),
    ]

    code: Annotated[
        str,
        Field(
            title="Error code",
            description="Code for this class of error",
            examples=["ServiceUnavailable"],
        ),
    ]

    message: Annotated[
        str,
        Field(
            title="Error message",
            description="Brief error messages",
            examples=["Short error message"],
        ),
    ]

    detail: Annotated[
        str | None,
        Field(
            title="Extended error message",
            description="Extended error message with additional detail",
            examples=["Some longer error message with details", None],
        ),
    ] = None

    def to_xml_model(self) -> ErrorSummary:
        """Convert to a Pydantic XML model."""
        return ErrorSummary(
            message=f"{self.code}: {self.message}",
            type=self.type,
            has_detail=self.detail is not None,
        )


class JobResult(BaseModel):
    """A single result from a job."""

    id: Annotated[
        str,
        Field(
            title="Result ID",
            description="Identifier for this result",
            examples=["image", "metadata"],
        ),
    ]

    url: Annotated[
        str,
        Field(
            title="Result URL",
            description="URL where the result is stored",
            examples=["s3://service-result-bucket/some-file"],
        ),
    ]

    size: Annotated[
        int | None,
        Field(
            title="Size of result",
            description="Size of the result in bytes if known",
            examples=[1238123, None],
        ),
    ] = None

    mime_type: Annotated[
        str | None,
        Field(
            title="MIME type of result",
            description="MIME type of the result if known",
            examples=["application/fits", "application/x-votable+xml", None],
        ),
    ] = None

    @classmethod
    def from_worker_result(cls, result: WorkerResult) -> Self:
        """Convert from the `~safir.arq.uws.WorkerResult` model."""
        return cls(
            id=result.result_id,
            url=result.url,
            size=result.size,
            mime_type=result.mime_type,
        )

    def to_xml_model(self) -> ResultReference:
        """Convert to a Pydantic XML model."""
        return ResultReference(
            id=self.id, size=self.size, mime_type=self.mime_type
        )


class SignedJobResult(JobResult):
    """A single result from the job with a signed URL.

    A `JobResult` is converted to a `SignedJobResult` before generating the
    response via templating or returning the URL as a redirect.
    """

    @override
    def to_xml_model(self) -> ResultReference:
        """Convert to a Pydantic XML model."""
        return ResultReference(
            id=self.id,
            type=None,
            href=self.url,
            size=self.size,
            mime_type=self.mime_type,
        )


class JobBase(BaseModel):
    """Fields common to all variations of the job record."""

    json_parameters: Annotated[
        dict[str, Any],
        Field(
            title="Job parameters",
            description=(
                "May be any JSON-serialized object. Stored opaquely and"
                " returned as part of the job record."
            ),
            examples=[
                {
                    "ids": ["data-id"],
                    "stencils": [
                        {
                            "type": "circle",
                            "center": [1.1, 2.1],
                            "radius": 0.001,
                        }
                    ],
                },
            ],
        ),
    ]

    run_id: Annotated[
        str | None,
        Field(
            title="Client-provided run ID",
            description=(
                "The run ID allows the client to add a unique identifier to"
                " all jobs that are part of a single operation, which may aid"
                " in tracing issues through a complex system or identifying"
                " which operation a job is part of"
            ),
            examples=["daily-2024-10-29"],
        ),
    ] = None

    destruction_time: Annotated[
        UtcDatetime,
        Field(
            title="Destruction time",
            description=(
                "At this time, the job will be aborted if it is still"
                " running, its results will be deleted, and it will either"
                " change phase to ARCHIVED or all record of the job will be"
                " discarded"
            ),
            examples=["2024-11-29T23:57:55+00:00"],
        ),
    ]

    execution_duration: Annotated[
        SecondsTimedelta | None,
        Field(
            title="Maximum execution duration",
            description=(
                "Allowed maximum execution duration. This is specified in"
                " elapsed wall clock time (not CPU time). If null, the"
                " execution time is unlimited. If the job runs for longer than"
                " this time period, it will be aborted."
            ),
        ),
        PlainSerializer(
            lambda t: int(t.total_seconds()) if t is not None else None,
            return_type=int,
        ),
    ] = None


class JobCreate(JobBase):
    """Information required to create a new UWS job (Wobbly format)."""


class SerializedJob(JobBase):
    """A single UWS job (Wobbly format)."""

    id: Annotated[
        str,
        Field(
            title="Job ID",
            description="Unique identifier of the job",
            examples=["47183"],
        ),
        BeforeValidator(lambda v: str(v) if isinstance(v, int) else v),
    ]

    service: Annotated[
        str,
        Field(
            title="Service",
            description="Service responsible for this job",
            examples=["vo-cutouts"],
        ),
    ]

    owner: Annotated[
        str,
        Field(
            title="Job owner",
            description="Identity of the owner of the job",
            examples=["someuser"],
        ),
    ]

    phase: Annotated[
        ExecutionPhase,
        Field(
            title="Execution phase",
            description="Current execution phase of the job",
            examples=[
                ExecutionPhase.PENDING,
                ExecutionPhase.EXECUTING,
                ExecutionPhase.COMPLETED,
            ],
        ),
    ]

    message_id: Annotated[
        str | None,
        Field(
            title="Work queue message ID",
            description=(
                "Internal message identifier for the work queuing system."
                " Only meaningful to the service that stored this ID."
            ),
            examples=["e621a175-e3bf-4a61-98d7-483cb5fb9ec2"],
        ),
    ] = None

    creation_time: Annotated[
        UtcDatetime,
        Field(
            title="Creation time",
            description="When the job was created",
            examples=["2024-10-29T23:57:55+00:00"],
        ),
    ]

    start_time: Annotated[
        UtcDatetime | None,
        Field(
            title="Start time",
            description="When the job started executing (if it has)",
            examples=["2024-10-30T00:00:21+00:00", None],
        ),
    ] = None

    end_time: Annotated[
        UtcDatetime | None,
        Field(
            title="End time",
            description="When the job stopped executing (if it has)",
            examples=["2024-10-30T00:08:45+00:00", None],
        ),
    ] = None

    quote: Annotated[
        UtcDatetime | None,
        Field(
            title="Expected completion time",
            description=(
                "Expected completion time of the job if it were started now,"
                " or null to indicate that the expected duration is not known."
                " If later than the destruction time, indicates that the job"
                " is not possible due to resource constraints."
            ),
        ),
    ] = None

    errors: Annotated[
        list[JobError],
        Field(
            title="Error", description="Error information if the job failed"
        ),
    ] = []

    results: Annotated[
        list[JobResult],
        Field(
            title="Job results",
            description="Results of the job, if it has finished",
        ),
    ] = []

    def to_job_description(self, base_url: str) -> ShortJobDescription:
        """Convert to the Pydantic XML model used for the summary of jobs.

        Parameters
        ----------
        base_url
            Base URL for the full jobs.
        """
        return ShortJobDescription(
            phase=self.phase,
            run_id=self.run_id,
            creation_time=self.creation_time,
            owner_id=self.owner,
            job_id=self.id,
            type=None,
            href=f"{base_url}/{self.id}",
        )


class Job[P: ParametersModel](SerializedJob):
    """A single UWS job with deserialized parameters."""

    parameters: Annotated[
        P,
        Field(
            title="Job parameters",
            description=(
                "Job parameters converted to their Pydantic model. Use"
                " ``json_parameters`` for the serialized form sent over"
                " the wire."
            ),
            exclude=True,
        ),
    ]

    @classmethod
    def from_serialized_job(
        cls, job: SerializedJob, parameters_type: type[P]
    ) -> Self:
        """Convert from a serialized job returned by Wobbly.

        Parameters
        ----------
        job
            Serialized job from Wobbly.
        parameters_type
            Model to use for job parameters.

        Returns
        -------
        Job
            Job with the correct parameters model.

        Raises
        ------
        pydantic.ValidationError
            Raised if the serialized parameters cannot be validated.
        """
        job_dict = job.model_dump()
        params = job_dict.get("json_parameters")
        if params:
            job_dict["parameters"] = parameters_type.model_validate(params)
        return cls.model_validate(job_dict)

    def to_serialized_job(self) -> SerializedJob:
        """Convert to a serialized job suitable for sending to Wobbly."""
        job = self.model_dump(mode="json")
        return SerializedJob.model_validate(job)

    def to_xml_model[S: JobSummary](self, job_summary_type: type[S]) -> S:
        """Convert to a Pydantic XML model.

        Parameters
        ----------
        job_summary_type
            XML model class for the job summary.

        Returns
        -------
        vo_models.uws.models.JobSummary
            XML model corresponding to this job.
        """
        results = None
        if self.results:
            results = Results(results=[r.to_xml_model() for r in self.results])
        duration = None
        if self.execution_duration:
            duration = int(self.execution_duration.total_seconds())
        error_summary = None
        if self.errors:
            error_summary = self.errors[0].to_xml_model()
        return job_summary_type(
            job_id=self.id,
            run_id=self.run_id,
            owner_id=self.owner,
            phase=self.phase,
            quote=self.quote,
            creation_time=self.creation_time,
            start_time=self.start_time,
            end_time=self.end_time,
            execution_duration=duration,
            destruction=self.destruction_time,
            parameters=self.parameters.to_xml_model(),
            results=results,
            error_summary=error_summary,
            version=UWSVersion.V1_1,
        )


class JobUpdateAborted(BaseModel):
    """Input model when aborting a job."""

    phase: Annotated[
        Literal[ExecutionPhase.ABORTED],
        Field(
            title="New phase",
            description="New phase of job",
            examples=[ExecutionPhase.ABORTED],
        ),
    ]


class JobUpdateCompleted(BaseModel):
    """Input model when marking a job as complete."""

    phase: Annotated[
        Literal[ExecutionPhase.COMPLETED],
        Field(
            title="New phase",
            description="New phase of job",
            examples=[ExecutionPhase.COMPLETED],
        ),
    ]

    results: Annotated[
        list[JobResult],
        Field(title="Job results", description="All the results of the job"),
    ]


class JobUpdateExecuting(BaseModel):
    """Input model when marking a job as executing."""

    phase: Annotated[
        Literal[ExecutionPhase.EXECUTING],
        Field(
            title="New phase",
            description="New phase of job",
            examples=[ExecutionPhase.EXECUTING],
        ),
    ]

    start_time: Annotated[
        UtcDatetime,
        Field(
            title="Start time",
            description="When the job started executing",
            examples=["2024-11-01T12:15:45+00:00"],
        ),
    ]


class JobUpdateError(BaseModel):
    """Input model when marking a job as failed."""

    phase: Annotated[
        Literal[ExecutionPhase.ERROR],
        Field(
            title="New phase",
            description="New phase of job",
            examples=[ExecutionPhase.ERROR],
        ),
    ]

    errors: Annotated[
        list[JobError],
        Field(
            title="Failure details",
            description="Job failure error message and details",
            min_length=1,
        ),
    ]


class JobUpdateQueued(BaseModel):
    """Input model when marking a job as queued."""

    phase: Annotated[
        Literal[ExecutionPhase.QUEUED],
        Field(
            title="New phase",
            description="New phase of job",
            examples=[ExecutionPhase.QUEUED],
        ),
    ]

    message_id: Annotated[
        str | None,
        Field(
            title="Queue message ID",
            description="Corresponding message within a job queuing system",
            examples=["4ce850a7-d877-4827-a3f6-f84534ec3fad"],
        ),
    ]


class JobUpdateMetadata(BaseModel):
    """Input model when updating job metadata."""

    phase: Annotated[
        None,
        Field(
            title="New phase", description="New phase of job", examples=[None]
        ),
    ] = None

    destruction_time: Annotated[
        UtcDatetime,
        Field(
            title="Destruction time",
            description=(
                "At this time, the job will be aborted if it is still"
                " running, its results will be deleted, and it will either"
                " change phase to ARCHIVED or all record of the job will be"
                " discarded"
            ),
            examples=["2024-11-29T23:57:55+00:00"],
        ),
    ]

    execution_duration: Annotated[
        SecondsTimedelta | None,
        Field(
            title="Maximum execution duration",
            description=(
                "Allowed maximum execution duration. This is specified in"
                " elapsed wall clock time (not CPU time). If null, the"
                " execution time is unlimited. If the job runs for longer than"
                " this time period, it will be aborted."
            ),
        ),
        PlainSerializer(
            lambda t: int(t.total_seconds()) if t else None,
            return_type=int | None,
        ),
    ]
